# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interface and implementation of clients talking to the token master."""

import abc
import random
import socket
import time

from thrift.transport import TSocket
from thrift.transport import TTransport
from thrift.protocol import TBinaryProtocol

from pinball.config.pinball_config import PinballConfig
from pinball.config.utils import get_log
from pinball.master.thrift_lib.ttypes import ArchiveRequest
from pinball.master.thrift_lib.ttypes import GroupRequest
from pinball.master.thrift_lib.ttypes import ModifyRequest
from pinball.master.thrift_lib.ttypes import QueryAndOwnRequest
from pinball.master.thrift_lib.ttypes import QueryRequest
from pinball.master.thrift_lib import TokenMasterService


__author__ = 'Pawel Garbacki, Mao Ye'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = ['Pawel Garbacki', 'Mao Ye']
__license__ = 'Apache'
__version__ = '2.0'


LOG = get_log('pinball.master.client')


class Client(object):
    """Interface of a client communicating with token master."""
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        # Mapping from request class to the end point that handles requests of
        # this type.
        self._request_to_end_point = None

    def call(self, request):
        return self._request_to_end_point[request.__class__](request)

    # For description of individual methods, see master.thrift.

    # TODO(pawel): remove these methods after replacing their invocations with
    # calls to call().

    def archive(self, request):
        return self.call(request)

    def group(self, request):
        return self.call(request)

    def modify(self, request):
        return self.call(request)

    def query(self, request):
        return self.call(request)

    def query_and_own(self, request):
        return self.call(request)


class LocalClient(Client):
    """Client communicating with master living in the same address space."""

    def __init__(self, master):
        super(LocalClient, self).__init__()
        self._master = master
        self._request_to_end_point = {
            ArchiveRequest: self._master.archive,
            GroupRequest: self._master.group,
            ModifyRequest: self._master.modify,
            QueryAndOwnRequest: self._master.query_and_own,
            QueryRequest: self._master.query}


class RemoteClient(Client):
    """Thrift client communicating with a remote master."""

    def __init__(self, host, port):
        super(RemoteClient, self).__init__()
        self._host = host
        self._port = port
        self._request_to_end_point = None
        self._connect()

    def __del__(self):
        self._transport.close()

    def _connect(self):
        backoff = PinballConfig.CLIENT_TIMEOUT_SEC
        for i in range(0, PinballConfig.CLIENT_CONNECT_ATTEMPTS):
            try:
                transport = TSocket.TSocket(self._host, self._port)
                transport.setTimeout(1000 * PinballConfig.CLIENT_TIMEOUT_SEC)
                self._transport = TTransport.TBufferedTransport(transport)
                protocol = TBinaryProtocol.TBinaryProtocol(self._transport)
                self._client = TokenMasterService.Client(protocol)
                self._request_to_end_point = {
                    ArchiveRequest: self._client.archive,
                    GroupRequest: self._client.group,
                    ModifyRequest: self._client.modify,
                    QueryAndOwnRequest: self._client.query_and_own,
                    QueryRequest: self._client.query}
                self._transport.open()
            except (TTransport.TTransportException, socket.timeout):
                LOG.exception('')
                if i == PinballConfig.CLIENT_CONNECT_ATTEMPTS - 1:
                    raise
                LOG.warning('failed during communication with master.  '
                            'Reconnecting %d / %d' % (
                    i + 1, PinballConfig.CLIENT_CONNECT_ATTEMPTS - 1))
                backoff *= 2
                time.sleep(self._get_backoff_time(backoff))

    def _get_backoff_time(self, backoff):
        """Get the backoff time for the client reconnecting to master."""
        return min(backoff * (1 + random.uniform(-0.5, 0.5)),
                   PinballConfig.MAX_BACKOFF_CLIENT_RECONNECT_SEC)

    def call(self, request):
        try:
            return super(RemoteClient, self).call(request)
        except (TTransport.TTransportException, socket.timeout, socket.error):
            LOG.exception('')
            self._connect()
            return super(RemoteClient, self).call(request)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Logic building data objects representing token aggregates."""
# TODO(pawel): move this file under workflow/
import collections
import operator
import pickle
import sys
import time

from pinball.config.utils import get_log
from pinball.config.utils import PinballException
from pinball.scheduler.overrun_policy import OverrunPolicy
from pinball.ui.data import JobData
from pinball.ui.data import JobExecutionData
from pinball.ui.data import Status
from pinball.ui.data import TokenData
from pinball.ui.data import TokenPathData
from pinball.ui.data import WorkflowScheduleData
from pinball.ui.data import WorkflowData
from pinball.ui.data import WorkflowInstanceData
from pinball.workflow import log_saver
from pinball.workflow.name import Name
from pinball.workflow.signaller import Signal


__author__ = 'Pawel Garbacki, Mao Ye'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = ['Pawel Garbacki', 'Mao Ye']
__license__ = 'Apache'
__version__ = '2.0'


LOG = get_log('pinball.ui.data_builder')


class DataBuilder(object):
    # TODO(pawel): change use_cache default to True after we gain enough
    # confidence that it is bug free.
    def __init__(self, store, use_cache=False):
        self._store = store
        self.use_cache = use_cache

    @staticmethod
    def _parse_job_token_name(token_name):
        name = Name.from_job_token_name(token_name)
        if name.workflow:
            return name
        return None

    @staticmethod
    def _get_instance_prefixes(token_names):
        prefixes = set()
        for token_name in token_names:
            name = DataBuilder._parse_job_token_name(token_name)
            if name:
                prefix = name.get_instance_prefix()
                assert prefix, ('Instance prefix not found in job token name '
                                '%s' % token_name)
                prefixes.add(prefix)
        return list(prefixes)

    @staticmethod
    def _job_status_from_job_token(job_token):
        """Extract job status from a job token.

        Args:
            job_token: The token to extract status from.
        Returns:
            Status of the job.
        """
        name = Name.from_job_token_name(job_token.name)
        job = pickle.loads(job_token.data)
        if not job.history:
            return Status.DISABLED if job.disabled else Status.NEVER_RUN
        last_execution_record = job.history[-1]
        if (name.job_state == Name.RUNNABLE_STATE and
                not last_execution_record.end_time):
            return Status.RUNNING
        if job.disabled:
            return Status.DISABLED
        if last_execution_record.exit_code != 0:
            return Status.FAILURE
        return Status.SUCCESS

    @staticmethod
    def _get_progress(execution_history, instance_start_time,
                      instance_end_time):
        """Extract info required to construct job execution progress bar.

        Args:
            execution_history: The list of execution records describing runs of
                this job.
            instance_start_time: The start time of the workflow instance that
                this job belongs to.
            instance_end_time: The end time of the workflow instance that this
                job belongs to or the current time if the instance did not yet
                finish.
        Returns:
            The list of (percentage, status) tuples where status describes
            the state that the job was in on the timeline of the instance,
            while the percentage tells for how long this status lasted as a
            percentage of the total instance run time.  Ordering of the tuples
            on the list follows the job state transitions so the job execution
            timeline can be reconstructed from its contents.
        """
        result = []
        instance_duration = instance_end_time - instance_start_time
        if instance_duration == 0:
            return result
        still_running = False
        first_start_time = None
        for record in execution_history:
            assert record.start_time, record
            if not first_start_time:
                first_start_time = record.start_time
            if record.end_time:
                execution_duration = record.end_time - record.start_time
                status = (Status.to_string(Status.SUCCESS) if
                          record.exit_code == 0 else
                          Status.to_string(Status.FAILURE))
            else:
                assert not still_running
                still_running = True
                execution_duration = instance_end_time - record.start_time
                status = Status.to_string(Status.RUNNING)
            percentage = int(100. * execution_duration / instance_duration)
            percentage = max(1, percentage)
            result.append((percentage, status))

        if first_start_time:
            duration = first_start_time - instance_start_time
        else:
            duration = instance_duration
        result.insert(0, (int(100. * duration / instance_duration), ''))
        return result

    @staticmethod
    def _job_data_from_job_token(job_token, instance_start_time,
                                 instance_end_time):
        """Extract job data from a job token.

        Args:
            job_token: The job token that should be converted to data.
            instance_start_time: The start time of the workflow instance that
                this job belongs to.
            instance_end_time: The end time of the workflow instance that this
                job belongs to or the current time if the instance did not yet
                finish.
        Returns:
            The job data extracted from the token.
        """
        status = DataBuilder._job_status_from_job_token(job_token)
        job = pickle.loads(job_token.data)
        if job.history:
            last_execution_record = job.history[-1]
            last_start_time = last_execution_record.start_time
            last_end_time = last_execution_record.end_time
        else:
            last_start_time = None
            last_end_time = None
        name = Name.from_job_token_name(job_token.name)
        progress = DataBuilder._get_progress(job.history,
                                             instance_start_time,
                                             instance_end_time)
        # TODO(mao): Change the status name from FAILURE to PENDING
        # if the condition is in pending status.
        return JobData(workflow=name.workflow,
                       instance=name.instance,
                       job=name.job,
                       job_type=job.__class__.__name__,
                       is_condition=job.IS_CONDITION,
                       info=job.info(),
                       inputs=job.inputs,
                       outputs=job.outputs,
                       emails=job.emails,
                       max_attempts=job.max_attempts,
                       retry_delay_sec=job.retry_delay_sec,
                       warn_timeout_sec=job.warn_timeout_sec,
                       abort_timeout_sec=job.abort_timeout_sec,
                       priority=job_token.priority,
                       status=status,
                       last_start_time=last_start_time,
                       last_end_time=last_end_time,
                       progress=progress)

    @staticmethod
    def _jobs_data_from_job_tokens(job_tokens):
        instance_start_time = sys.maxint
        instance_end_time = 0
        now = time.time()
        for job_token in job_tokens:
            job = pickle.loads(job_token.data)
            if job.history:
                start_time = job.history[0].start_time
                instance_start_time = min(instance_start_time, start_time)
                end_time = (job.history[-1].end_time
                            if job.history[-1].end_time else now)
                instance_end_time = max(instance_end_time, end_time)
        if instance_start_time == sys.maxint:
            instance_start_time = now
        if instance_end_time == 0:
            instance_end_time = now
        result = []
        for job_token in job_tokens:
            job_data = DataBuilder._job_data_from_job_token(
                job_token, instance_start_time, instance_end_time)
            result.append(job_data)
        return result

    @staticmethod
    def _job_data_less_than(job_data1, job_data2):
        """A comparator for job data objects.

        Jobs are sorted based on the execution time.  Finished jobs sort on the
        end time.  A running job sorts after a finished job.  Running jobs sort
        on the start time.  Jobs that never run sort on qualified name.
        """
        if (not job_data1.last_start and not job_data2.last_start and
                not job_data1.last_end and not job_data2.last_end):
            # neither job ever run
            name1 = Name(workflow=job_data1.workflow,
                         instance=job_data1.instance,
                         job_state=Name.WAITING_STATE,
                         job=job_data1.job).get_job_token_name()
            name2 = Name(workflow=job_data2.workflow,
                         instance=job_data2.instance,
                         job_state=Name.WAITING_STATE,
                         job=job_data2.job).get_job_token_name()
            return name1 < name2

        if not job_data1.last_start and job_data2.last_start:
            # job1 never run, job2 did
            return True
        if job_data1.last_start and not job_data2.last_start:
            # job2 never run, job1 did
            return False
        if not job_data1.last_end and not job_data2.last_end:
            # both jobs are running
            return job_data1.last_start < job_data2.last_start
        if not job_data1.last_end and job_data2.last_end:
            # only job1 is running
            return False
        if job_data1.last_end and not job_data2.last_end:
            # only job2 is running
            return True
        # both jobs run in the past but neither is running now
        return job_data1.last_end < job_data2.last_end

    def _get_signal(self, workflow, instance, action, active):
        signal = Signal.action_to_string(action)
        name = Name(workflow=workflow, instance=instance, signal=signal)
        signal_token_name = name.get_signal_token_name()
        if active:
            tokens = self._store.read_active_tokens(
                name_prefix=signal_token_name)
        else:
            tokens = self._store.read_archived_tokens(
                name_prefix=signal_token_name)
        if not tokens:
            return None
        assert len(tokens) == 1
        assert tokens[0].name == signal_token_name
        return pickle.loads(tokens[0].data)

    def _instance_data_from_job_tokens(self, job_tokens):
        """Extract instance data from job tokens in that instance.

        Args:
            job_tokens: Job tokens that belong to a single workflow instance.
        Returns:
            Workflow data describing the workflow instance identified by the
            input job tokens.
        """
        assert job_tokens
        start_time = time.time()
        end_time = 0
        failed = False
        for job_token in job_tokens:
            job = pickle.loads(job_token.data)
            if job.history:
                first_execution_record = job.history[0]
                if (first_execution_record.start_time and
                        first_execution_record.start_time < start_time):
                    start_time = first_execution_record.start_time
                last_execution_record = job.history[-1]
                if not last_execution_record.end_time:
                    end_time = sys.maxint
                else:
                    if last_execution_record.end_time > end_time:
                        end_time = last_execution_record.end_time
                    if (not job.disabled and
                            last_execution_record.exit_code != 0):
                        failed = True
        if not job_tokens:
            is_active = False
        else:
            is_active = True
            job_name = job_tokens[0].name
            archived_tokens = self._store.read_archived_tokens(
                name_prefix=job_name)
            for token in archived_tokens:
                if token.name == job_name:
                    is_active = False
                    break

        name = Name.from_job_token_name(job_tokens[0].name)
        is_scheduled_for_archive = False
        abort_signal = None
        if is_active:
            archive_signal = self._get_signal(name.workflow,
                                              name.instance,
                                              Signal.ARCHIVE,
                                              True)
            is_scheduled_for_archive = (archive_signal and
                                        Signal.TIMESTAMP_ATTR in
                                        archive_signal.attributes)
        else:
            abort_signal = self._get_signal(name.workflow,
                                            name.instance,
                                            Signal.ABORT,
                                            False)
        if abort_signal:
            status = Status.ABORTED
            if end_time == 0:
                # This can happen only if all jobs have an empty history.
                timestamp = abort_signal.attributes.get(Signal.TIMESTAMP_ATTR)
                start_time = timestamp
                end_time = timestamp
        elif (end_time == 0 or end_time == sys.maxint or
              (is_active and not is_scheduled_for_archive)):
            status = Status.RUNNING
            end_time = None
        elif failed:
            status = Status.FAILURE
        else:
            status = Status.SUCCESS
        return WorkflowInstanceData(name.workflow,
                                    name.instance,
                                    status,
                                    start_time,
                                    end_time)

    def _instances_data_from_job_tokens(self, job_tokens):
        """Extract instance data from job tokens.

        Args:
            job_tokens: Job tokens, potentially from different instances.
        Returns:
            List of workflow instance data defined by input job tokens.
        """
        result = []
        job_tokens_per_instance = collections.defaultdict(list)
        for job_token in job_tokens:
            name = Name.from_job_token_name(job_token.name)
            job_tokens_per_instance[name.get_instance_prefix()].append(
                job_token)
        for _, tokens in job_tokens_per_instance.items():
            result.append(
                self._instance_data_from_job_tokens(tokens))
        return result

    @staticmethod
    def _workflow_data_from_instances_data(instances):
        """Extract workflow data from its instances data.

        'last_instance' is defined as:
            - the one that started most recently if there is running instance
            - the one that finished most recently if this is no running instance
        (finish means non-running, could be: success, failed or aborted)

        Args:
            instances: Workflow instances to extract workflow data from.
            Instances must belong to the same workflow.
        Returns:
            Workflow data defined by input instances.
        """
        assert instances

        running_instance_number = 0
        last_instance = None
        # workflow property (property of last_instance)
        wf_start_time = 0
        wf_end_time = 0
        wf_status = Status.RUNNING

        for instance in instances:
            # handle a running instance
            if not instance.end_time:
                wf_status = Status.RUNNING
                running_instance_number += 1
                # the previous last workflow is non-running or it's running but
                # starts earlier that this instance
                if wf_end_time or instance.start_time > wf_start_time:
                    wf_start_time = instance.start_time
                    last_instance = instance.instance
                wf_end_time = None
            # no running instance is ever found yet
            elif wf_end_time is not None:
                # reuse start time if we lost the actual end time
                this_end_time = instance.start_time \
                    if instance.end_time == sys.maxint else instance.end_time
                if this_end_time > wf_end_time:
                    wf_start_time = instance.start_time
                    wf_end_time = this_end_time
                    wf_status = instance.status
                    last_instance = instance.instance

        return WorkflowData(instances[0].workflow,
                            wf_status,
                            wf_start_time,
                            wf_end_time,
                            last_instance,
                            running_instance_number)

    def _workflows_data_from_instances_data(self, instances):
        """Extract workflows data from instances data.

        Args:
            instances: Workflows instances to extract workflows data from.
        Returns:
            List of workflows data defined by input instances data.
        """
        instances_per_workflow = collections.defaultdict(list)
        for instance in instances:
            instances_per_workflow[instance.workflow].append(instance)
        result = []
        for _, instances in instances_per_workflow.items():
            result.append(
                DataBuilder._workflow_data_from_instances_data(instances))
        return result

    def _workflows_data_from_job_tokens(self, job_tokens):
        """Extract workflows data from job tokens.

        Args:
            job_tokens: Job tokens to extract workflows data from.
        Returns:
            List of workflows data defined by input job tokens.
        """
        assert job_tokens
        instances_data = \
            self._instances_data_from_job_tokens(job_tokens)
        return self._workflows_data_from_instances_data(instances_data)

    def _get_job_tokens(self, workflow=None, instance=None, job_state=None,
                        job=None):
        """Extract job tokens from the store.

        Args:
            workflow: The name of the workflow whose jobs we are interested in.
            instance: The name of the instance whose jobs we are interested in.
            job_state: The state of the jobs we are interested in.
            job: The name of the job we are interested in.
        Returns:
            List of jobs matching the specification.
        """
        name = Name(workflow=workflow, instance=instance, job_state=job_state,
                    job=job)
        if name.job:
            prefix = name.get_job_token_name()
        elif name.job_state:
            prefix = name.get_job_state_prefix()
        elif name.instance:
            prefix = name.get_job_prefix()
        elif name.workflow:
            prefix = name.get_workflow_prefix()
        else:
            prefix = ''
        tokens = self._store.read_tokens(name_prefix=prefix)
        result = []
        for token in tokens:
            token_name = Name.from_job_token_name(token.name)
            if token_name.get_job_token_name():
                # This is a job token.
                if not job or job == token_name.job:
                    # We matched the prefix so if we are looking for a specific
                    # job, its names must match exactly.
                    result.append(token)
        return result

    def _get_job(self, workflow, instance, job):
        """Get job definition from the store.

        Args:
            workflow: The name of the job workflow.
            instance: The name of the job instance.
            job: The name of the job.
        Returns:
            Matching job definition.
        """
        job_token = None
        for state in [Name.WAITING_STATE, Name.RUNNABLE_STATE]:
            # There is a small chance that a job will move from one state to
            # another in between the iterations and the method will return None
            # although the job actually exists.
            job_tokens = self._get_job_tokens(workflow=workflow,
                                              instance=instance,
                                              job_state=state,
                                              job=job)
            assert len(job_tokens) <= 1
            if len(job_tokens) == 1:
                job_token = job_tokens[0]
                break
        if job_token:
            return pickle.loads(job_token.data)
        return None

    def _get_jobs(self, workflow, job):
        """Get job definitions from the store across all workflow instances.

        Args:
            workflow: The name of the job workflow.
            instance: The name of the job instance.
            job: The name of the job.
        Returns:
            Matching job definition.
        """
        name = Name(workflow=workflow)
        name_prefix = name.get_workflow_prefix()
        # This is a bit hacky since we bypass the Name module where all the
        # token naming logic is supposed to be located.
        # TODO(pawel): extend the Name module to support abstractions needed
        # here.
        name_infix = '/job/'
        name_suffix = '/%s' % job
        job_tokens = self._store.read_tokens(name_prefix=name_prefix,
                                             name_infix=name_infix,
                                             name_suffix=name_suffix)
        result = []
        for job_token in job_tokens:
            job_record = pickle.loads(job_token.data)
            result.append(job_record)
        return result

    def _get_workflows_using_cache(self):
        """Get workflows, preferably fetching instances from the cache.

        As a side effect, archived instances that do not exist in the cache
        will be added to the cache.

        Returns:
            List of workflows data.
        """
        instances_token_names = self._store.read_token_names(
            name_prefix=Name.WORKFLOW_PREFIX)
        instances_prefixes = DataBuilder._get_instance_prefixes(
            instances_token_names)
        instances = []
        for prefix in instances_prefixes:
            name = Name.from_instance_prefix(prefix)
            assert name.workflow and name.instance, (
                'Expected instance prefix, found %s' % prefix)
            instances.append(self.get_instance(name.workflow, name.instance))
        return self._workflows_data_from_instances_data(instances)

    def get_workflows(self):
        """Get all workflows data from the store.

        Returns:
            List of workflows data.
        """
        if self.use_cache:
            return self._get_workflows_using_cache()
        all_tokens = self._get_job_tokens()
        if not all_tokens:
            return []
        return self._workflows_data_from_job_tokens(all_tokens)

    def get_workflow(self, workflow):
        """Get workflow data from the store.

        Args:
            workflow: The name of the workflow whose instance should be
                retrieved.
        Returns:
            Workflow data or None if it was not found.
        """
        instances = self.get_instances(workflow)
        if not instances:
            return None
        return DataBuilder._workflow_data_from_instances_data(instances)

    def _get_instances_using_cache(self, workflow):
        """Get workflow instances, preferably from the cache.

        As a side effect, archived instances that do not exist in the cache
        will be added to the cache.

        Args:
            workflow: The name of the workflow whose instances we are
                interested in.
        Returns:
            List of instances for the given workflow.
        """
        name = Name(workflow=workflow)
        workflow_prefix = name.get_workflow_prefix()
        workflow_token_names = self._store.read_token_names(
            name_prefix=workflow_prefix)
        instances_prefixes = DataBuilder._get_instance_prefixes(
            workflow_token_names)
        result = []
        for prefix in instances_prefixes:
            name = Name.from_instance_prefix(prefix)
            assert name.workflow and name.instance, (
                'Expected instance prefix, found %s' % prefix)
            result.append(self.get_instance(name.workflow, name.instance))
        return result

    def get_instances(self, workflow):
        """Get from the store workflow instances data for a given workflow.

        Args:
            workflow: The name of the workflow whose instances we are
                interested in.
        Returns:
            List of instances for the given workflow.
        """
        if self.use_cache:
            return self._get_instances_using_cache(workflow)
        workflow_tokens = self._get_job_tokens(workflow=workflow)
        if not workflow_tokens:
            return []
        return self._instances_data_from_job_tokens(workflow_tokens)

    def get_latest_instance(self, workflow):
        """Get latest workflow instance.

        Args:
            workflow: The name of the workflow whose instance we are
                interested in.
        Returns:
            The workflow instance or None if it was not found.
        """
        instances = self.get_instances(workflow)
        return max(instances, key=operator.attrgetter('start_time'))

    def _get_instance_no_cache(self, workflow, instance):
        """Get workflow instance, bypass the cache.

        Args:
            workflow: The name of the workflow whose instance we are
                interested in.
            instance: The instance we are interested in.
        Returns:
            The workflow instance or None if it was not found.
        """
        instance_tokens = self._get_job_tokens(workflow=workflow,
                                               instance=instance)
        if not instance_tokens:
            return None
        return self._instance_data_from_job_tokens(instance_tokens)

    def _get_instance_using_cache(self, workflow, instance):
        """Get workflow instance, preferably from the cache.

        As a side effect, if the instance is archived and it does not exist in
        the cache, it will be added to the cache.

        Args:
            workflow: The name of the workflow whose instance we are
                interested in.
            instance: The instance we are interested in.
        Returns:
            The workflow instance or None if it was not found.
        """
        name = Name(workflow=workflow, instance=instance)
        instance_prefix = name.get_instance_prefix()
        data = self._store.get_cached_data(instance_prefix)
        if data:
            instance_data = pickle.loads(data)
        else:
            # Cache only archived instances.
            if self._store.read_archived_token_names(
                    name_prefix=instance_prefix):
                # The ordering of operations is important.  We need to make
                # sure that we add to the cache instance data constructed from
                # the archived tokens.
                instance_data = self._get_instance_no_cache(workflow, instance)
                self._store.set_cached_data(instance_prefix,
                                            pickle.dumps(instance_data))
            else:
                instance_data = self._get_instance_no_cache(workflow, instance)
        return instance_data

    def get_instance(self, workflow, instance):
        """Get workflow instance data from the store.

        Args:
            workflow: The name of the workflow whose instance we are
                interested in.
            instance: The instance we are interested in.
        Returns:
            The workflow instance or None if it was not found.
        """
        if self.use_cache:
            return self._get_instance_using_cache(workflow, instance)
        return self._get_instance_no_cache(workflow, instance)

    def get_jobs(self, workflow, instance):
        """Get from the store jobs data for a given workflow and instance.

        Args:
            workflow: The name of the workflow whose jobs we are interested in.
            instance: The name of the instance whose jobs we are interested in.
        Returns:
            List of jobs for the given workflow and instance.
        """
        instance_tokens = self._get_job_tokens(workflow=workflow,
                                               instance=instance)
        if not instance_tokens:
            return []
        return DataBuilder._jobs_data_from_job_tokens(instance_tokens)

    @staticmethod
    def _execution_record_to_execution_data(workflow, job, execution,
                                            execution_record):
        return JobExecutionData(
            workflow=workflow,
            instance=execution_record.instance,
            job=job,
            execution=execution,
            info=execution_record.info,
            exit_code=execution_record.exit_code,
            cleanup_exit_code=execution_record.cleanup_exit_code,
            start_time=execution_record.start_time,
            end_time=execution_record.end_time,
            properties=execution_record.properties,
            logs=execution_record.logs)

    def get_executions(self, workflow, instance, job):
        """Get from the store executions data for a given job.

        Args:
            workflow: The name of the workflow whose executions we are
                interested in.
            instance: The name of the instance whose executions we are
                interested in.
            job: The name of the job whose executions we are interested in.
        Returns:
            List of executions for the given workflow, instance, and job.
        """
        job_record = self._get_job(workflow, instance, job)
        result = []
        if not job_record:
            return result
        i = 0
        for execution_record in job_record.history:
            result.append(DataBuilder._execution_record_to_execution_data(
                workflow, job, i, execution_record))
            i += 1
        return result

    def get_executions_across_instances(self, workflow, job):
        """Get from the store executions data for a given job across instances.

        Args:
            workflow: The name of the workflow whose executions we are
                interested in.
            job: The name of the job whose executions we are interested in.
        Returns:
            List of executions for the given workflow and job.
        """
        job_records = self._get_jobs(workflow, job)
        executions_data = []
        if not job_records:
            return executions_data
        for job_record in job_records:
            i = 0
            for execution_record in job_record.history:
                executions_data.append(
                    DataBuilder._execution_record_to_execution_data(
                        workflow, job, i, execution_record))
                i += 1
        sorted_executions_data = sorted(executions_data,
                                        key=lambda execution_data:
                                        (execution_data.start_time,
                                         execution_data.instance,
                                         execution_data.execution))
        # The same execution can be present in multiple tokens if the the
        # workflow was retried.  Therefore we need to de-dup the records.
        result = []
        last_execution_data = None
        for execution_data in sorted_executions_data:
            if (not last_execution_data or
                    execution_data.instance != last_execution_data.instance or
                    execution_data.execution != last_execution_data.execution):
                result.append(execution_data)
                last_execution_data = execution_data
        return result

    def get_execution(self, workflow, instance, job, execution):
        """Get execution data from the store.

        Args:
            workflow: The name of the workflow whose execution we are
                interested in.
            instance: The name of the instance whose execution we are
                interested in.
            job: The name of the job whose execution we are interested in.
            execution: The execution number.
        Returns:
            The execution matching the input parameters.
        """
        job_record = self._get_job(workflow, instance, job)
        if not job_record or len(job_record.history) <= execution:
            return None
        execution_record = job_record.history[execution]
        return JobExecutionData(
            workflow=workflow,
            instance=instance,
            job=job,
            execution=execution,
            info=execution_record.info,
            exit_code=execution_record.exit_code,
            cleanup_exit_code=execution_record.cleanup_exit_code,
            start_time=execution_record.start_time,
            end_time=execution_record.end_time,
            properties=execution_record.properties,
            logs=execution_record.logs)

    def get_file_content(self, workflow, instance, job, execution, log_type):
        """Get content of a given execution log file.

        Args:
            workflow: The name of the workflow whose log we are interested in.
            instance: The name of the instance whose log we are interested in.
            job: The name of the job whose log we are interested in.
            execution: The execution whose log we are interested in.
            log_type: The type of the log we are interested in.
        Returns:
            The content of the matching log file.
        """
        job_record = self._get_job(workflow, instance, job)
        if not job_record:
            return ''
        execution_record = job_record.history[execution]
        if log_type in execution_record.logs:
            file_name = execution_record.logs[log_type]
            f = log_saver.FileLogSaver.from_path(file_name)
            try:
                f.open('r')
                return f.read()
            except:
                LOG.exception('')
            finally:
                # If the log is local file, then we need to close the file.
                if f and not file_name.startswith('s3n://'):
                    f.close()
        else:
            return ''

    def get_schedules(self):
        """Get all workflow schedules data from the store."""
        tokens = self._store.read_tokens(
            name_prefix=Name.WORKFLOW_SCHEDULE_PREFIX)
        result = []
        for token in tokens:
            schedule = pickle.loads(token.data)
            overrun_policy_help = (
                OverrunPolicy.get_help(schedule.overrun_policy))
            result.append(WorkflowScheduleData(
                next_run_time=schedule.next_run_time,
                recurrence_seconds=schedule.recurrence_seconds,
                overrun_policy=schedule.overrun_policy,
                overrun_policy_help=overrun_policy_help,
                workflow=schedule.workflow,
                parser_params=schedule.parser_params,
                emails=schedule.emails,
                max_running_instances=schedule.max_running_instances))
        return result

    def get_schedule(self, workflow):
        """Get workflow schedule data from the store.

        Args:
            workflow: The name of the workflow whose schedule should be
                retrieved.
        Returns:
            The workflow schedule or None if it was not found.
        """
        name = Name(workflow=workflow)
        schedule_token_name = name.get_workflow_schedule_token_name()
        tokens = self._store.read_tokens(name_prefix=schedule_token_name)
        if tokens:
            for token in tokens:
                if token.name == schedule_token_name:
                    schedule = pickle.loads(token.data)
                    overrun_policy_help = OverrunPolicy.get_help(
                        schedule.overrun_policy)
                    return WorkflowScheduleData(
                        next_run_time=schedule.next_run_time,
                        recurrence_seconds=schedule.recurrence_seconds,
                        overrun_policy=schedule.overrun_policy,
                        overrun_policy_help=overrun_policy_help,
                        workflow=schedule.workflow,
                        parser_params=schedule.parser_params,
                        emails=schedule.emails,
                        max_running_instances=schedule.max_running_instances)
        return None

    def get_token_paths(self, path):
        """Get token paths data from the store.

        Args:
            path: The path is the name prefix of the parent whose direct
                children should be returned.
        Returns:
            List of direct path descendants of the parent.
        """
        if not path.startswith(Name.DELIMITER):
            raise PinballException('incorrectly formatted path %s' % path)
        # TODO(pawel): this is a bit inefficient as it may load names of quite
        # a few tokens into the memory.
        token_names = self._store.read_token_names(name_prefix=path)
        counts = collections.defaultdict(int)
        path_len = len(path)
        for token_name in token_names:
            index = token_name.find(Name.DELIMITER, path_len)
            if index == -1:
                index = len(token_name)
            else:
                index += 1
            group = token_name[:index]
            counts[group] += 1
        result = []
        for path, count in counts.items():
            result.append(TokenPathData(path, count))
        return result

    def get_token(self, name):
        """Get token data from the store.

        Args:
            path: The name of the token to retrieve.
        Returns:
            The token or None if it was not found.
        """
        tokens = self._store.read_tokens(name)
        token = None
        for token in tokens:
            if token.name == name:
                break
        if not token or token.name != name:
            raise PinballException("didn't find any tokens with name %s" %
                                   name)
        return TokenData(name=token.name,
                         version=token.version,
                         owner=token.owner,
                         expiration_time=token.expirationTime,
                         priority=token.priority,
                         data=token.data)

    def is_signal_set(self, workflow, instance, action):
        """Check if a signal is set.

        Args:
            workflow: The workflow whose signal should be checked.  If None,
                signals at the global level are checked.
            instance: The workflow instance whose signal should be checked.  If
                not None, a matching workflow name must be provided.
                If None, signals at the workflow and the global level are
                checked.
            action: The signal action to check.
        Returns:
            True iff the signal exists in the specified context.
        """
        for (workflow_name, instance_name) in [(workflow, instance),
                                               (workflow, None),
                                               (None, None)]:
            name = Name(workflow=workflow_name, instance=instance_name,
                        signal=Signal.action_to_string(action))
            token_name = name.get_signal_token_name()
            tokens = self._store.read_tokens(token_name)
            assert len(tokens) <= 1
            if tokens:
                return True
        return False



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Token data represents the entities stored in the token data attribute."""
import abc


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class TokenData(object):
    """Parent class for entities stored in token data field."""
    __metaclass__ = abc.ABCMeta

    @property
    def _COMPATIBILITY_ATTRIBUTES(self):
        """Return attributes that should be set if missing.

        The set of object attributes may change between versions.  We need the
        ability to deserialize (unpickle) older versions of an object into new
        versions of object implementation.  To this end, we offer the ability
        to fill attributes missing in the serialized (old) objects with default
        values.

        This property may be overridden in subclasses.

        Returns:
            Dictionary of attribute names and their default values to be filled
            in if missing during unpickling.
        """
        return {}

    def __setstate__(self, state):
        self.__dict__ = state
        for attribute, default in self._COMPATIBILITY_ATTRIBUTES.items():
            if attribute not in self.__dict__:
                self.__dict__[attribute] = default



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Signal the worker.

Signal structures stored in tokens provide a way to communicate with workflow
workers.  Workers periodically query for signal tokens and execute actions
stored in those tokens.

Signal tokens may be placed at different levels of the workflow hierarchy,
i.e., the top level, the workflow level, and the instance level.  Location of
the signal token defines its scope.  E.g., a DRAIN token posted to
/workflow/__SIGNAL__/ drains all workflows in the system, a token located in
/workflow/<workflow>/__SIGNAL__/ drains instances of workflow <workflow> only,
and a token in /workflow/<workflow>/<instance>/__SIGNAL__/ will drain only
instance <instance> of workflow <workflow>.

TODO(pawel): certain types of signals should be restricted to specific levels.
E.g., EXIT signal should be posted only at the top level, and ABORT token at an
instance level.
"""
import pickle
import time

from pinball.config.pinball_config import PinballConfig
from pinball.master.thrift_lib.ttypes import ModifyRequest
from pinball.master.thrift_lib.ttypes import Query
from pinball.master.thrift_lib.ttypes import QueryRequest
from pinball.master.thrift_lib.ttypes import Token
from pinball.persistence.token_data import TokenData
from pinball.workflow.name import Name


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class Signal(TokenData):
    """A signal communicates actions to workflow workers."""
    # Action types.
    DRAIN, ABORT, ARCHIVE, EXIT = range(4)

    _ACTION_NAMES = {
        # Finish currently running jobs but do not start new ones.
        DRAIN: 'DRAIN',
        # Abort all running jobs, do not start new ones.
        ABORT: 'ABORT',
        # Archive the workflow if there are no runnable jobs.
        ARCHIVE: 'ARCHIVE',
        # Shut down the worker.  It only makes sense to define this action at
        # the top level.
        EXIT: 'EXIT'
    }

    GENERATION_ATTR, TIMESTAMP_ATTR = range(2)

    _ATTRIBUTE_NAMES = {
        GENERATION_ATTR: 'GENERATION',
        TIMESTAMP_ATTR: 'TIMESTAMP'
    }

    def __init__(self, action, attributes=None):
        self.action = action
        self.attributes = {} if attributes is None else attributes

    @property
    def _COMPATIBILITY_ATTRIBUTES(self):
        return {
            'attributes': {}
        }

    @staticmethod
    def action_to_string(action):
        return Signal._ACTION_NAMES[action]

    @staticmethod
    def attribute_to_string(attribute):
        return Signal._ATTRIBUTE_NAMES[attribute]

    def __str__(self):
        action_str = Signal.action_to_string(self.action)
        return 'Signal(action=%s, attributes=%s)' % (action_str,
                                                     self.attributes)

    def __repr__(self):
        return self.__str__()


class Signaller(object):
    """Signaller delivers and retrieves signals."""
    def __init__(self, client, workflow=None, instance=None):
        assert workflow or not instance
        self._client = client
        self._workflow = workflow
        self._instance = instance
        self._signals = {}  # mapping from action to signal
        self._refresh_actions()

    def _dedup_actions(self, signal_tokens):
        """Remove action duplicates from a list of signal tokens.

        Duplicate actions are possible because the same action may be signalled
        at different levels of the hierarchy.  If it happens, we arbitrarily
        choose one of the duplicates.

        This method refreshes signals stored internally from the list of input
        tokens.

        Args:
            signal_tokens: The signal tokens to dedup.
        """
        self._signals = {}
        for signal_token in signal_tokens:
            signal = pickle.loads(signal_token.data)
            self._signals[signal.action] = signal

    def _refresh_actions(self):
        """Reload actions from the master."""
        request = QueryRequest(queries=[])
        name = Name()

        top_query = Query()
        top_query.namePrefix = name.get_signal_prefix()
        request.queries.append(top_query)

        if self._workflow:
            workflow_query = Query()
            name.workflow = self._workflow
            workflow_query.namePrefix = name.get_signal_prefix()
            request.queries.append(workflow_query)

        if self._instance:
            instance_query = Query()
            name.instance = self._instance
            instance_query.namePrefix = name.get_signal_prefix()
            request.queries.append(instance_query)

        response = self._client.query(request)
        signal_tokens = []
        for tokens in response.tokens:
            signal_tokens.extend(tokens)

        self._dedup_actions(signal_tokens)

    def _get_signal_token(self, action):
        """Retrieve signal for a specific action from the master.

        Args:
            action: The action to get signal for.
        Returns:
            The signal token if found, otherwise None.
        """
        request = QueryRequest(queries=[])
        name = Name(workflow=self._workflow,
                    instance=self._instance,
                    signal=Signal.action_to_string(action))

        query = Query()
        query.namePrefix = name.get_signal_token_name()
        request.queries.append(query)

        response = self._client.query(request)
        assert len(response.tokens) == 1
        tokens = response.tokens[0]
        if not tokens:
            return None
        assert len(tokens) == 1
        return tokens[0]

    def is_signal_present(self, action):
        """Check if a signal is set.

        Consult local signal storage only.  No calls to the master are made.

        Args:
            action: The action we are interested in.
        Returns:
            True iff a signal for the specific action exists.
        """
        return action in self._signals

    def is_action_set(self, action):
        """Check if a signal for a specific action exists.

        Consult local signal storage only.  No calls to the master are made.

        This method is different from is_signal_present as it interprets signal
        attribute set to determine if the signal is relevant in the context of
        the current program.  An example is the EXIT token whose applicability
        depends on the generation.

        Args:
            action: The action we are interested in.
        Returns:
            True iff a signal for the action exists and the action is relevant
            in the local context.
        """
        signal = self._signals.get(action)
        if not signal:
            return False
        if action == Signal.EXIT:
            generation = signal.attributes.get(Signal.GENERATION_ATTR)
            if generation and generation <= PinballConfig.GENERATION:
                return False
        return signal is not None

    def _send_request(self, request):
        """Send modify request to the master.

        If the request fails, locally stored signals get refreshed from the
        master.

        Args:
            request: The modify request to send.
        Returns:
            True iff the request has succeeded.
        """
        try:
            self._client.modify(request)
        except:
            # This can happen if someone concurrently posts the same signal
            # token.
            self._refresh_actions()
            return False
        else:
            return True

    def set_action(self, action):
        """Send a signal with a specific action to the master.

        Local signal store gets updated with the new action if it is
        successfully submitted to the master.  If the communication with the
        master fails, locally stored signals get refreshed.

        Args:
            action: The action to set.
        """
        attributes = {}
        if action == Signal.ABORT:
            attributes[Signal.TIMESTAMP_ATTR] = time.time()
        elif action == Signal.EXIT:
            attributes[Signal.GENERATION_ATTR] = PinballConfig.GENERATION
        signal = self._signals.get(action)
        if signal and signal.attributes == attributes:
            return
        # A signal with the same action but different data may already exist
        # in the master.
        signal_token = self._get_signal_token(action)
        if not signal_token:
            name = Name(workflow=self._workflow, instance=self._instance,
                        signal=Signal.action_to_string(action))
            signal_token = Token(name=name.get_signal_token_name())
        signal = Signal(action, attributes)
        signal_token.data = pickle.dumps(signal)
        request = ModifyRequest(updates=[signal_token])
        if self._send_request(request):
            self._signals[action] = signal

    def remove_action(self, action):
        """Remove signal with a given action from the master.

        Args:
            action: The action to remove.
        """
        if not self.is_signal_present(action):
            return
        signal_token = self._get_signal_token(action)
        if signal_token:
            request = ModifyRequest(deletes=[signal_token])
            self._client.modify(request)
        del self._signals[action]

    def get_attribute(self, action, attribute):
        """Get the value of a specific attribute.

        Args:
            action: The action whose attribute we are interested in.
            attribute: The attribute we are interested in.
        Returns:
            The value of the attribute or None if that attribute was not found.
        """
        signal = self._signals.get(action)
        if not signal:
            return None
        return signal.attributes.get(attribute)

    def set_attribute_if_missing(self, action, attribute, value):
        """Set an attribute value unless that attribute is already set.

        Args:
            action: The action whose attribute should be set.
            attribute: The attribute to set.
            value: The attribute value to set.
        Returns:
            True iff the attribute value was set.  Return False if the signal
            is not set.
        """
        if self.get_attribute(action, attribute) is not None:
            return False
        signal_token = self._get_signal_token(action)
        if not signal_token:
            return False
        signal = pickle.loads(signal_token.data)
        self._signals[action] = signal
        if self.get_attribute(action, attribute) is not None:
            return False
        signal.attributes[attribute] = value
        signal_token.data = pickle.dumps(signal)
        request = ModifyRequest(updates=[signal_token])
        if self._send_request(request):
            self._signals[action] = signal
            return True
        return False



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pinball_ext.common import utils
from pinball_ext.job.basic_jobs import ClusterJob


__author__ = 'Changshu Liu, Mohammad Shahangian, Mao Ye'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


LOG = utils.get_logger('pinball_ext.job.hadoop_jobs')


class HadoopJob(ClusterJob):
    """Base class for actual Hadoop jobs.

    App jar and lib jars are configured in executor, please see
    Executor.run_hadoop_job() for detailed info.

    Derived class should at least override _get_class_name() to specify what's
    the main Java class to execute. It can also optionally override _setup() to
    config the follow parameters to further tune the job config:
    - self.jobconf_args
    - self.extra_jars
    - self.extra_arguments

    self.params derived from JobBase will also be passed as job's extra
    arguments (together with self.extra_arguments).
    """
    def __init__(self, params, settings=None):
        super(HadoopJob, self).__init__(params, settings)
        self.jobconf_args = {}
        self.extra_arguments = []
        self.extra_jars = []

    def _get_class_name(self):
        raise NotImplementedError('No class name specified for this Hadoop Job')

    def _execute(self):
        param_args = ['-%s %s' % (k, v) for k, v in self.params.iteritems()]

        self._job_output, self._job_stderr, self._job_ids = \
            self.executor.run_hadoop_job(
                self._get_class_name(),
                jobconf_args=self.jobconf_args,
                extra_args=param_args + self.extra_arguments,
                extra_jars=self.extra_jars)

        LOG.info('Dump job output ...')
        for line in self._job_output:
            LOG.info('\t'.join(line))

    def __str__(self):
        return '(%s): (%s) - (%s)' % (self.job_name,
                                      self.params,
                                      self.jobconf_args)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Various decorator classes."""

import sys
import time

from pinball_ext.common.utils import get_logger


__author__ = 'Mao Ye'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


LOG = get_logger('pinball_ext.common.decorators.retry')


def retry(ExceptionToCheck, tries=4, delay=3, backoff=2, logger=LOG,
          sleep_func=time.sleep, max_delay=sys.maxint):
    """Retry calling the decorated function using an exponential backoff.

    http://www.saltycrane.com/blog/2009/11/trying-out-retry-decorator-python/
    original from: http://wiki.python.org/moin/PythonDecoratorLibrary#Retry

    Args:
        ExceptionToCheck: exception to check. May be a tuple of
            exceptions to check.
        tries: an integer, number of times to try (not retry) before
            giving up.
        delay: an integer, initial delay between retries in seconds.
        backoff: an integer, backoff multiplier e.g. value of 2 will
            double the delay each retry
        logger: logging.Logger instance, logger to use. By default,
            we use ``logging.Logger.log``; if None is explicitly specified by
            the caller, ``print`` is used.
        sleep_func: the sleep function to be used for waiting between
            retries. By default, it is ``time.sleep``,
            but it could also be gevent.sleep if we are using this with
            gevent.
        max_delay: the max number of seconds to wait between retries.

    Returns:
        Decorator function.
    """
    def deco_retry(f):

        def f_retry(*args, **kwargs):
            mtries, mdelay = tries, delay
            while mtries > 1:
                try:
                    return f(*args, **kwargs)
                except ExceptionToCheck, e:
                    if logger:
                        # Don't evaluate the message string as logger module
                        # uses lazy evaluation: assembles the string only if it
                        # is going to be logged at the current logging level.
                        logger.warning(
                            "%s, Retrying in %d seconds...", e, mdelay)
                    else:
                        print "%s, Retrying in %d seconds..." % (
                            str(e), mdelay)
                    sleep_func(mdelay)
                    mtries -= 1
                    mdelay *= backoff
                    # Don't wait more than max_delay allowed
                    if mdelay > max_delay:
                        mdelay = max_delay
            return f(*args, **kwargs)

        return f_retry  # True decorator.

    return deco_retry



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Analyzer allows to manipulate workflow instance tokens.

Only archived workflow instances can be analyzed.  Tokens of an archived
workflow instance are immutable so we don't need to worry that the workflow
state will change while we manipulate the graph.
"""
import pickle

from pinball.config.pinball_config import PinballConfig
from pinball.master.thrift_lib.ttypes import Query
from pinball.master.thrift_lib.ttypes import QueryRequest
from pinball.master.thrift_lib.ttypes import Token
from pinball.parser.config_parser import ParserCaller
from pinball.parser.utils import load_parser_with_caller
from pinball.workflow.name import Name
from pinball.workflow.event import Event
from pinball.workflow.utils import load_path


__author__ = 'Pawel Garbacki, Mao Ye'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class Analyzer(object):
    def __init__(self, workflow, instance):
        self._workflow = workflow
        self._instance = instance
        self._jobs = {}
        self._existing_events = {}
        self._new_events = {}
        self._job_priorities = {}

    @staticmethod
    def from_store(store, workflow, instance):
        """Construct analyzer from tokens in a store.

        Args:
            store: The store to read tokens from.
            workflow: The workflow whose tokens should be read.
            instance: The instance whose tokens should be read.

        Returns:
            Analyzer initialized with tokens read from the store.
        """
        analyzer = Analyzer(workflow, instance)
        analyzer._read_tokens_from_store(store)
        return analyzer

    @staticmethod
    def from_parser_params(workflow):
        """Construct analyzer from tokens of a workflow according
        to the parser params configuration.

        Args:
            workflow: The workflow whose tokens should be read.

        Returns:
            Analyzer initialized with tokens read from the config.
        """
        analyzer = Analyzer(workflow, None)
        analyzer._read_tokens_from_parser_params()
        return analyzer

    @staticmethod
    def from_client(client, workflow, instance):
        """Construct analyzer from tokens of a workflow in the master.

        Args:
            client: The client connected to the master.
            workflow: The workflow whose tokens should be read.
            instance: The instance whose tokens should be read.

        Returns:
            Analyzer initialized with tokens read from the client.
        """
        analyzer = Analyzer(workflow, instance)
        analyzer._read_tokens_from_client(client)
        return analyzer

    def _filter_job_tokens(self, tokens):
        """Filter out all tokens which are not job tokens.

        Args:
            tokens: The tokens to filter.
        """
        for token in tokens:
            name = Name.from_job_token_name(token.name)
            if not self._instance and name.instance:
                self._instance = name.instance
            if name.job:
                job = pickle.loads(token.data)
                self._jobs[job.name] = job
                self._job_priorities[job.name] = token.priority

    def _filter_event_tokens(self, tokens):
        """Filter out all tokens which are not event tokens.

        Args:
            tokens: The tokens to filter.
        """
        for token in tokens:
            name = Name.from_event_token_name(token.name)
            if not self._instance and name.instance:
                self._instance = name.instance
            if name.event:
                event = pickle.loads(token.data)
                self._existing_events[token.name] = event

    def _read_tokens_from_store(self, store):
        """Read archived job tokens from the store.

        Args:
            store: The store to read tokens from.
        """
        name = Name(workflow=self._workflow, instance=self._instance)
        tokens = store.read_archived_tokens(
            name_prefix=name.get_instance_prefix())
        self._filter_job_tokens(tokens)

    def _read_tokens_from_parser_params(self):
        """Read archived job tokens from the PinballConfig.PARSER_PARAMS.
        """
        config_parser = load_parser_with_caller(PinballConfig.PARSER,
                                                PinballConfig.PARSER_PARAMS,
                                                ParserCaller.ANALYZER)
        tokens = config_parser.get_workflow_tokens(self._workflow)
        self._filter_job_tokens(tokens)

    def _read_tokens_from_client(self, client):
        """Read archived job tokens from the client.

        Args:
            client: The client to read tokens from.
        """
        name = Name(workflow=self._workflow, instance=self._instance)
        query = Query(namePrefix=name.get_workflow_prefix())
        request = QueryRequest(queries=[query])
        response = client.query(request)
        assert len(response.tokens) == 1
        tokens = response.tokens[0]
        self._filter_job_tokens(tokens)
        self._filter_event_tokens(tokens)

    def get_tokens(self):
        """Export all internally stored tokens.

        Returns:
            The list of tokens after all transformations performed by the
            analyzer.
        """
        result = []
        for job in self._jobs.values():
            name = Name(workflow=self._workflow, instance=self._instance,
                        job_state=Name.WAITING_STATE, job=job.name)
            data = pickle.dumps(job)
            token = Token(name=name.get_job_token_name(),
                          priority=self._job_priorities[job.name],
                          data=data)
            result.append(token)
        result.extend(self.get_new_event_tokens())
        return result

    def get_new_event_tokens(self):
        """Export new event tokens.

        Returns:
            The list of new event tokens after all transformations performed by
            the analyzer.
        """
        result = []
        for event_name, event in self._new_events.items():
            data = pickle.dumps(event)
            token = Token(name=event_name, data=data)
            result.append(token)
        return result

    def _find_descendants(self, job_name):
        """Find direct and indirect descendants of a job.

        Args:
            job_name: The name of the job whose descendants should be computed.

        Returns:
            The set of job descendants.
        """
        def _dfs(current, visited):
            if not current in visited:
                visited.add(current)
                for child in current.outputs:
                    _dfs(self._jobs[child], visited)
        visited = set()
        _dfs(self._jobs[job_name], visited)
        result = set()
        for job in visited:
            result.add(job.name)
        return result

    def _generate_missing_events(self, job_names):
        """Generate external events required to run all jobs in a set.

        For a set of jobs (a subset of all jobs in the workflow), produce
        events satisfying upstream dependencies external to that set.  E.g.,
        for job dependency structure like this:

        A1  A2
         | /
        B1  B2
         |
        C1  C2
         | /
        D1

        and job_names = (C1, D1) we would generate events satisfying the
        following deps: B1->C1, C2->D1.

        Args:
            job_names: The set of job names whose external deps are to be
                satisfied by the generated events.
        """
        input_prefixes = set()
        for job_name in job_names:
            job = self._jobs[job_name]
            for job_input in job.inputs:
                if job_input not in job_names:
                    name = Name(workflow=self._workflow,
                                instance=self._instance,
                                job=job_name,
                                input_name=job_input,
                                event='poison_%d' % len(input_prefixes))
                    input_prefix = name.get_input_prefix()
                    if input_prefix not in input_prefixes:
                        input_prefixes.add(input_prefix)
                        event_token_name = name.get_event_token_name()
                        if not event_token_name in self._existing_events:
                            self._new_events[
                                name.get_event_token_name()] = Event(
                                    'analyzer')

    def poison(self, roots):
        """Poison the workflow instance at specified root jobs.

        Poisoning is a process of choosing a subset of jobs to run based on
        the dependencies.  Only jobs that depend directly or indirectly on any
        job in the root set will be executed.  Poisoning will generate events
        for dependencies which are not satisfied by the roots or their
        dependents.

        Args:
            roots: The list of job names which are roots of the poisoning.
        """
        descendants = set()
        for root in roots:
            jobs = self._find_descendants(root)
            descendants = descendants.union(jobs)
        self._generate_missing_events(descendants)

    def change_instance(self, instance):
        """Move all tokens to a specific instance."""
        self._instance = instance
        new_events = {}
        for event_name, event in self._new_events.items():
            name = Name.from_event_token_name(event_name)
            name.instance = instance
            new_events[name.get_event_token_name()] = event
        self._new_events = new_events

    def clear_job_histories(self):
        """Remove histories from all job tokens."""
        for job in self._jobs.values():
            job.history = []



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines an example workflow."""
from datetime import datetime
from datetime import timedelta

from pinball.config.pinball_config import PinballConfig
from pinball_ext.examples.jobs import FINAL_JOB
from pinball_ext.workflow.config import JobConfig
from pinball_ext.workflow.config import WorkflowConfig
from pinball_ext.workflow.config import ScheduleConfig
from pinball_ext.job_templates import JobTemplate


WORKFLOWS = {
    'example_workflow': WorkflowConfig(
        jobs={
            'parent_job': JobConfig(JobTemplate('ExamplePythonJob'), []),
            'child_job': JobConfig(JobTemplate('ExampleCommandJob'), ['parent_job'])
        },
        final_job_config=JobConfig(FINAL_JOB),
        schedule=ScheduleConfig(recurrence=timedelta(days=1),
                                reference_timestamp=datetime(
                                year=2015, month=2, day=1, second=1)),
        notify_emails='%s' % PinballConfig.DEFAULT_EMAIL)
}



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Data structures defining data formats send over to Web clients."""
# TODO(pawel): move this file under workflow/
import copy
import time

from pinball.config.utils import PinballException
from pinball.config.utils import timestamp_to_str
from pinball.config.utils import token_data_to_str
from pinball.scheduler.overrun_policy import OverrunPolicy


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class Status(object):
    UNKNOWN, NEVER_RUN, RUNNING, FAILURE, SUCCESS, DISABLED, ABORTED, CONDITION_PENDING = range(8)

    _STATUS_NAMES = {UNKNOWN: 'UNKNOWN',
                     NEVER_RUN: 'NEVER_RUN',
                     RUNNING: 'RUNNING',
                     SUCCESS: 'SUCCESS',
                     FAILURE: 'FAILURE',
                     DISABLED: 'DISABLED',
                     ABORTED: 'ABORTED',
                     CONDITION_PENDING: 'CONDITION_PENDING'}

    COLORS = {UNKNOWN: '#000000',
              NEVER_RUN: '#dddddd',
              RUNNING: '#66ccff',
              SUCCESS: '#89e894',
              FAILURE: '#ff3333',
              DISABLED: '#ffffff',
              ABORTED: '#ff3333',
              CONDITION_PENDING: '#ffcc00'}

    @staticmethod
    def to_string(status):
        return Status._STATUS_NAMES[status]

    @staticmethod
    def from_string(status_name):
        for status, name in Status._STATUS_NAMES.items():
            if name == status_name:
                return status
        raise PinballException('Unknown status %s' % status_name)


def _get_run_time(start_time, end_time):
    if not start_time:
        return None
    if not end_time:
        end_time = time.time()
    return int(end_time - start_time)


class WorkflowData(object):
    def __init__(self, workflow, status=None, last_start_time=None,
                 last_end_time=None, last_instance=None,
                 running_instance_number=None):
        self.workflow = workflow
        self.status = status
        self.last_start_time = last_start_time
        self.last_end_time = last_end_time
        self.last_instance = last_instance
        self.running_instance_number = running_instance_number

    def format(self):
        result = copy.copy(self.__dict__)
        result['status'] = Status.to_string(self.status)
        if self.last_start_time:
            result['last_start_time'] = timestamp_to_str(self.last_start_time)
        if self.last_end_time:
            result['last_end_time'] = timestamp_to_str(self.last_end_time)
        # TODO(pawel): run_time can be inferred from start and end time.  To
        # make it easier to parse in the UI, we should send start and end time
        # in seconds.
        result['run_time'] = _get_run_time(self.last_start_time,
                                           self.last_end_time)
        result['running_instance_number'] = str(self.running_instance_number)
        return result


class WorkflowInstanceData(object):
    def __init__(self, workflow, instance, status=None, start_time=None,
                 end_time=None):
        self.workflow = workflow
        self.instance = instance
        self.status = status
        self.start_time = start_time
        self.end_time = end_time

    def format(self):
        result = copy.copy(self.__dict__)
        result['status'] = Status.to_string(self.status)
        if self.start_time:
            result['start_time'] = timestamp_to_str(self.start_time)
        if self.end_time:
            result['end_time'] = timestamp_to_str(self.end_time)
        result['run_time'] = _get_run_time(self.start_time, self.end_time)
        return result


class JobData(object):
    def __init__(self, workflow, instance, job, job_type, is_condition, info,
                 inputs, outputs, emails=None, max_attempts=1,
                 retry_delay_sec=0, warn_timeout_sec=None,
                 abort_timeout_sec=None, priority=None, status=None,
                 last_start_time=None, last_end_time=None, progress=None):
        self.workflow = workflow
        self.instance = instance
        self.job = job
        self.job_type = job_type
        self.is_condition = is_condition
        self.info = info
        self.inputs = inputs
        self.outputs = outputs
        self.emails = emails if emails is not None else []
        self.max_attempts = max_attempts
        assert self.max_attempts > 0
        self.retry_delay_sec = retry_delay_sec
        self.warn_timeout_sec = warn_timeout_sec
        self.abort_timeout_sec = abort_timeout_sec
        self.priority = priority
        self.status = status
        self.last_start_time = last_start_time
        self.last_end_time = last_end_time
        self.progress = progress if progress else []

    def format(self):
        result = copy.copy(self.__dict__)
        result['status'] = Status.to_string(self.status)
        if self.last_start_time:
            result['last_start_time'] = timestamp_to_str(self.last_start_time)
        if self.last_end_time:
            result['last_end_time'] = timestamp_to_str(self.last_end_time)
        result['run_time'] = _get_run_time(self.last_start_time,
                                           self.last_end_time)
        return result


class JobExecutionData(object):
    def __init__(self, workflow, instance, job, execution, info,
                 exit_code=None, cleanup_exit_code=None, start_time=None,
                 end_time=None, properties=None, logs=None):
        self.workflow = workflow
        self.instance = instance
        self.job = job
        self.execution = execution
        self.info = info
        self.exit_code = exit_code
        self.cleanup_exit_code = cleanup_exit_code
        self.start_time = start_time
        self.end_time = end_time
        self.properties = properties if properties is not None else {}
        self.logs = logs if logs is not None else []

    def format(self):
        result = copy.copy(self.__dict__)
        if self.start_time:
            result['start_time'] = timestamp_to_str(self.start_time)
        if self.end_time:
            result['end_time'] = timestamp_to_str(self.end_time)
        result['run_time'] = _get_run_time(self.start_time, self.end_time)
        return result


class WorkflowScheduleData(object):
    def __init__(self, next_run_time, recurrence_seconds, overrun_policy,
                 overrun_policy_help, workflow, parser_params, emails,
                 max_running_instances):
        self.next_run_time = next_run_time
        self.recurrence_seconds = recurrence_seconds
        self.overrun_policy = overrun_policy
        self.overrun_policy_help = overrun_policy_help
        self.workflow = workflow
        self.parser_params = parser_params
        self.emails = emails
        self.max_running_instances = max_running_instances

    def format(self):
        result = copy.copy(self.__dict__)
        result['next_run_time'] = timestamp_to_str(self.next_run_time)
        result['overrun_policy'] = OverrunPolicy.to_string(self.overrun_policy)
        result['max_running_instances'] = str(self.max_running_instances)
        result['parser_params'] = str(self.parser_params)
        return result


class TokenPathData(object):
    def __init__(self, path, count):
        self.path = path
        self.count = count

    def format(self):
        return copy.copy(self.__dict__)


class TokenData(object):
    def __init__(self, version, name, owner, expiration_time, priority, data):
        self.version = version
        self.name = name
        self.owner = owner
        self.expiration_time = expiration_time
        self.priority = priority
        self.data = data

    def format(self):
        result = copy.copy(self.__dict__)
        if not self.owner:
            result['owner'] = ''
        result['expiration_time'] = timestamp_to_str(self.expiration_time)
        result['data'] = token_data_to_str(self.data)
        return result



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pinball configuration object"""

import socket
import yaml


__author__ = 'Mao Ye'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class PinballConfig(object):

    # GENERATION should be incremented with each change affecting any of the
    # pinball servers (master, workers, etc.)
    GENERATION = 1

    # Maximum number of running instances allowed for a workflow
    DEFAULT_MAX_WORKFLOW_RUNNING_INSTANCES = 3

    # Master and client configuration.
    MASTER_HOST = socket.gethostname()
    MASTER_PORT = 9090
    CLIENT_CONNECT_ATTEMPTS = 10
    CLIENT_TIMEOUT_SEC = 3 * 60

    # Number of workers
    WORKERS = 50
    # A delay between starting individual workers.  We space starting new
    # workers to prevent overwhelming the master.
    WORKER_CREATION_SLEEP_INTERVAL_SEC = 10 * 60
    WORKER_POLL_TIME_SEC = 5 * 60

    # Maximum backoff time for client reconnect to master
    MAX_BACKOFF_CLIENT_RECONNECT_SEC = 20 * 60

    # Application configuration for Pinball persistence layer.
    INSTALLED_APPS = (
        'pinball.authentication',
        'pinball.persistence',
        'django.contrib.admin',
        'django.contrib.auth',
        'django.contrib.contenttypes',
        'django.contrib.sessions',
        'django.contrib.sites',
        'django.contrib.messages',
        'django.contrib.staticfiles',
    )

    DEFAULT_MIDDLEWARE_CLASSES = (
        'django.contrib.sessions.middleware.SessionMiddleware',
        'django.contrib.messages.middleware.MessageMiddleware',
        'django.middleware.common.CommonMiddleware',
        'django.middleware.csrf.CsrfViewMiddleware',
    )

    MIDDLEWARE_CLASSES = DEFAULT_MIDDLEWARE_CLASSES

    # The UI_HOST is a VPN-friendly link for use to link Pinball logs in email.
    UI_HOST = None
    UI_PORT = 8080

    # UI server configuration.
    import os
    import pinball.ui as project_module
    ROOT_URLCONF = 'pinball.ui.urls'
    PROJECT_ROOT = os.path.dirname(os.path.realpath(project_module.__file__))
    TEMPLATE_DIRS = (
        os.path.join(PROJECT_ROOT, 'templates'),
    )
    STATICFILES_DIRS = (
        os.path.join(PROJECT_ROOT, 'assets'),
    )
    STATIC_ROOT = os.path.join(PROJECT_ROOT, 'static')
    ADMINS = (
    )
    MANAGERS = ADMINS
    STATIC_URL = '/static/'
    from django.conf.global_settings import TEMPLATE_CONTEXT_PROCESSORS as TCP
    TEMPLATE_CONTEXT_PROCESSORS = TCP + (
        'django.core.context_processors.request',
    )

    # Configuration for storing job log files in a remote storage.
    # By default, we disable logging to the remote storage.
    # If pinball ui and workers are run on different machines,
    # this parameter should be set.
    JOB_LOG_PATH_PREFIX = 'pinball_job_logs/'
    LOCAL_LOGS_DIR_PREFIX = '/tmp/'
    LOCAL_LOGS_DIR = LOCAL_LOGS_DIR_PREFIX + JOB_LOG_PATH_PREFIX
    S3_LOGS_DIR_PREFIX = None
    S3_LOGS_DIR = None

    # Use the following two settings when DEFAULT_S3_LOGS_DIR is set.
    AWS_ACCESS_KEY_ID = None
    AWS_SECRET_ACCESS_KEY = None

    # Pinball db settings
    DATABASES = {
        # Django requires a default database.
        'default': {
            'ENGINE': 'django.db.backends.mysql',
            'NAME': 'pinball',
            'USER': 'flipper',
            'PASSWORD': 'flipper123',
            'HOST': '127.0.0.1',
            'PORT': '3306',
        },
        'pinball.persistence': {
            'ENGINE': 'django.db.backends.mysql',
            'NAME': 'pinball',
            'USER': 'flipper',
            'PASSWORD': 'flipper123',
            'HOST': '127.0.0.1',
            'PORT': '3306',
        }
    }

    # Authentication configuration.
    AUTHENTICATION_DOMAINS = []
    GOOGLE_CLIENT_ID = None
    GOOGLE_CLIENT_SECRET = None

    # Django configuration parameters.
    # Secret key must be set in Django >= 1.5.
    # This was used for encryption in the legacy version
    SECRET_KEY = '0ma^86f1%7z5dbw57a5spwplab64oh-3wtq+05@akoto_&c0!!'
    AES_CBC_KEY = '\x91@\xc8\xa5I\xb2\xf2u\xb6e\xac\xc0\x03\x86U\x0b\xd4\x13' +\
        '\xca\x9b)\x9d\xc4t\x81s\x95\x89\x16\x9e\x8f\xa7'
    HMAC_KEY = '\xaf[\x0b\xb0\xa0\x8e\xbe\xa1\xb9:%\x9c\xb4yB\xf2n{\xa9"\x96' +\
        '\x83\xe7\t\xca\xb3T\xd9l\x8e\x08\xed'
    CRYPTO_VERSION = 1


    # Configuration for self-serving ETL creation through Pinball
    GITHUB_API_URI = '/api/v3/'
    GITHUB_HTTP_TIMEOUT_SEC = 30
    GITHUB_OAUTH_TOKEN = None
    GITHUB_SERVER = None
    USER = 'data'

    # Application configuration.
    PARSER = 'pinball_ext.workflow.parser.PyWorkflowParser'
    PARSER_PARAMS = {}

    # Configuration for the default email in pinball.
    # We use it as default sender of email service right now.
    DEFAULT_EMAIL = None

    # Debugging should be disabled for long running Django apps to prevent caching
    # of queries leading to ever growing memory consumption.
    DEBUG = True

    ALLOWED_HOSTS = ['*']

    # SMTP Configuration
    SMTP_HOST = 'localhost'
    SMTP_PORT = 25
    SMTP_USER = ''
    SMTP_PASS = ''
    SMTP_SSL  = False

    @staticmethod
    def parse(config_file):
        # Assume the config file is "yaml" file
        yaml_file = None
        try:
            yaml_file = open(config_file, 'r')
            config_data = yaml.load(yaml_file)
        except:
            raise
        finally:
            if yaml_file:
                yaml_file.close()

        for key, value in config_data.iteritems():
            PinballConfig.update_config(key.upper(), value)
        PinballConfig._check_configs()
        PinballConfig._post_process_configs()

    @classmethod
    def update_config(cls_obj, config_name, config_value):
        setattr(cls_obj, config_name, config_value)

    @staticmethod
    def _post_process_configs():
        PinballConfig.LOCAL_LOGS_DIR = \
            PinballConfig.LOCAL_LOGS_DIR_PREFIX \
            + PinballConfig.JOB_LOG_PATH_PREFIX
        if PinballConfig.S3_LOGS_DIR_PREFIX is not None:
            PinballConfig.S3_LOGS_DIR = \
                PinballConfig.S3_LOGS_DIR_PREFIX \
                + PinballConfig.JOB_LOG_PATH_PREFIX
        if PinballConfig.AUTHENTICATION_DOMAINS:
            PinballConfig.MIDDLEWARE_CLASSES = \
                ('pinball.authentication.oauth2.RequireLogin',) \
                + PinballConfig.DEFAULT_MIDDLEWARE_CLASSES[:1] \
                + ('django.contrib.auth.middleware.AuthenticationMiddleware',) \
                + PinballConfig.DEFAULT_MIDDLEWARE_CLASSES[2:]

    @staticmethod
    def _check_configs():
        assert (PinballConfig.DEFAULT_EMAIL is not None), \
            'Please specify the default_email!'
        assert PinballConfig.PARSER_PARAMS, \
            'Please configure parser_params!'
        if PinballConfig.AUTHENTICATION_DOMAINS:
            assert (PinballConfig.GOOGLE_CLIENT_ID is not None), \
                'Please specify the google_client_id!'
            assert (PinballConfig.GOOGLE_CLIENT_SECRET is not None), \
                'Please specify the google_client_secret!'
            assert (PinballConfig.SECRET_KEY is not None), \
                'Please specify the secrent_key for django oauth!'
            assert (PinballConfig.AES_CBC_KEY is not None), \
                'Please specify the aes_cbc_key for django oauth!'
            assert (PinballConfig.HMAC_KEY is not None), \
                'Please specify the hmac_key for django oauth!'
            assert (PinballConfig.CRYPTO_VERSION), \
                'Please specify the crypto_version for django oauth!'



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Validation tests for workflow parser."""
from datetime import datetime
from datetime import timedelta
import pickle
import unittest

import mock
from pinball.scheduler.schedule import OverrunPolicy
from pinball.scheduler.schedule import WorkflowSchedule
from pinball_ext.workflow.config import JobConfig
from pinball_ext.workflow.config import ScheduleConfig
from pinball_ext.workflow.config import WorkflowConfig
from pinball_ext.job_templates import CommandJobTemplate
from pinball_ext.workflow.parser import _is_name_qualified
from pinball_ext.workflow.parser import _get_qualified_name
from pinball_ext.workflow.parser import JobDef
from pinball_ext.workflow.parser import WorkflowDef


__author__ = 'Pawel Garbacki, Mao Ye, Jooseong Kim'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class ToolsTestCase(unittest.TestCase):
    def test_is_name_qualified(self):
        self.assertFalse(_is_name_qualified('workflow_job'))
        self.assertTrue(_is_name_qualified('workflow.job'))

    def test_get_qualified_name(self):
        self.assertEqual('workflow.job', _get_qualified_name('workflow',
                                                             'job'))


class JobDefTestCase(unittest.TestCase):
    def setUp(self):
        self.workflow = WorkflowDef('some_workflow', 'some_schedule',
                                    'some_emails')
        template = CommandJobTemplate('some_template', 'some_command')
        output_template = CommandJobTemplate('output_template', 'output_command', priority=30.0)
        self.job = JobDef('some_job', template, self.workflow)
        self.output_job = JobDef('output_job', output_template, self.workflow)
        self.job.outputs = [self.output_job]

    def test_to_string(self):
        """Test conversion of job to string."""

        self.assertEqual('job:some_workflow.some_job, deps:[]', str(self.job))

    def test_get_qualified_name(self):
        """Test construction of qualified job names."""

        self.assertEqual('some_workflow.some_job',
                         self.job.get_qualified_name())

    def test_get_canonical_name(self):
        """Test construction of canonical job names."""

        self.assertEqual('some_other_workflow_some_workflow_some_job',
                         self.job.get_canonical_name('some_other_workflow'))

    def test_compute_score(self):
        """Test the compute_score method of jobs"""
        self.assertEqual(30.0, self.output_job.compute_score())
        self.assertEqual(31.0, self.job.compute_score())


class WorkflowDefTestCase(unittest.TestCase):
    def setUp(self):
        # 1370354400 = 06/04/2012 2:00p UTC
        self.next_run_time = 1370354400
        # 1370289240 = 06/03/2013 7:53p UTC
        self.previous_run_time = 1370289240
        self.workflow_schedule = WorkflowSchedule(next_run_time=self.next_run_time,
                                                  recurrence_seconds=24 * 60 * 60,
                                                  workflow='some_workflow',
                                                  emails=['some_email'])
        self.workflow = WorkflowDef('some_workflow',
                                    self.workflow_schedule,
                                    'some_emails')
        self.job1 = JobDef('some_job_1', CommandJobTemplate('some_template1', 'some_command1'),
                           self.workflow)
        self.workflow.add_job(self.job1)
        self.job2 = JobDef('some_job_2', CommandJobTemplate('some_template2', 'some_command2'),
                           self.workflow)
        self.workflow.add_job(self.job2)
        self.job2.add_dep(self.job1)
        self.job3 = JobDef('some_job_3', CommandJobTemplate('some_template3', 'some_command3'),
                           self.workflow)
        self.workflow.add_job(self.job3)
        self.job4 = JobDef('some_job_4', CommandJobTemplate('some_template4', 'some_command4'),
                           self.workflow)
        self.workflow.add_job(self.job4)
        self.job4.add_dep(self.job3)

    def _add_final_job(self):
        template = CommandJobTemplate('some_template', 'some_command')
        final_job = JobDef('final_job', template, self.workflow)
        self.workflow.add_job(final_job)
        final_job.add_dep(self.job2)
        final_job.add_dep(self.job4)

    def _add_external_deps(self):
        other_workflow = WorkflowDef('some_other_workflow', 'some_schedule',
                                     'some_emails')
        template = CommandJobTemplate('some_template', 'some_command')
        parent_external_job = JobDef('parent_external_job', template,
                                     other_workflow)
        other_workflow.add_job(parent_external_job)
        external_job = JobDef('external_job', template, other_workflow)
        external_job.add_dep(parent_external_job)
        other_workflow.add_job(external_job)
        self.job1.add_dep(external_job)

    def test_to_string(self):
        """Test conversion of job to string."""

        self.assertEqual(
            "workflow:some_workflow, schedule:WorkflowSchedule(next_run_time=2013-06-04 "
            "14:00:00 UTC, recurrence=1 day, 0:00:00, overrun_policy=SKIP, "
            "parser_params={'caller': 'schedule'}, "
            "workflow=some_workflow, email=['some_email'], max_running_instances=3), "
            "notify_emails:some_emails, jobs:{'some_job_3': "
            "job:some_workflow.some_job_3, deps:[], 'some_job_2': "
            "job:some_workflow.some_job_2, deps:[some_workflow.some_job_1], "
            "'some_job_1': job:some_workflow.some_job_1, deps:[], "
            "'some_job_4': job:some_workflow.some_job_4, "
            "deps:[some_workflow.some_job_3]}", str(self.workflow))

    def test_get_leaf_jobs(self):
        """Test retrieval of leaf jobs."""

        leaf_jobs = self.workflow.get_leaf_jobs()
        leaf_jobs.sort()
        expected_jobs = [self.job2, self.job4]
        expected_jobs.sort()
        self.assertEqual(expected_jobs, leaf_jobs)

    def test_verify(self):
        """Test workflow verification."""

        # TODO(pawel): figure out why putting this at the top level does not
        # work.
        from pinball_ext.workflow.parser import WorkflowVerificationException

        # Raise an exception if the workflow does not have a final job.
        self.assertRaises(WorkflowVerificationException, self.workflow.verify)

        # Verify that the workflow has a final job and is acyclic.
        self._add_final_job()
        self.workflow.verify()

        # Raise an exception if the workflow has a cycle.
        self.job1.add_dep(self.job2)
        self.assertRaises(Exception, self.workflow.verify)

    @mock.patch('time.time')
    def test_get_schedule_token(self, time_mock):
        # the value must be a float
        time_mock.return_value = 1.0 * self.previous_run_time
        self._add_final_job()
        token = self.workflow.get_schedule_token()
        self.assertEqual('/schedule/workflow/some_workflow', token.name)
        schedule = pickle.loads(token.data)
        self.assertEqual(self.next_run_time, schedule.next_run_time)
        self.assertEqual(24 * 60 * 60, schedule.recurrence_seconds)
        self.assertEqual(OverrunPolicy.SKIP, schedule.overrun_policy)
        self.assertEqual('some_workflow', schedule.workflow)

    def test_get_workflow_tokens(self):
        self._add_final_job()
        self._add_external_deps()
        tokens = self.workflow.get_workflow_tokens()
        # There should be 5 (4 local and 1 external) workflow job tokens, 1
        # final job token, and 2 top-level job event tokens.
        self.assertEqual(5 + 1 + 2, len(tokens))


class PyWorkflowParserTestCase(unittest.TestCase):
    def test_get_schedule_token(self):
        # Due to a wired interaction between tests caused by dynamic module
        # imports, the following import statement cannot be placed at the top
        # level.  E.g., see
        # http://stackoverflow.com/questions/9722343/python-super-behavior-not-dependable
        from pinball_ext.workflow.parser import PyWorkflowParser
        params =\
            {'workflows_config':
                 'tests.pinball_ext.workflow.parser_test.WORKFLOWS'}
        py_workflow_parser = PyWorkflowParser(params)
        schedule_token = py_workflow_parser.get_schedule_token('some_workflow')
        self.assertEqual('/schedule/workflow/some_workflow',
                         schedule_token.name)

    def test_get_workflow_tokens(self):
        from pinball_ext.workflow.parser import PyWorkflowParser
        params =\
            {'workflows_config':
                 'tests.pinball_ext.workflow.parser_test.WORKFLOWS'}
        py_workflow_parser = PyWorkflowParser(params)
        workflow_tokens = py_workflow_parser.get_workflow_tokens(
            'some_workflow')
        # Two job tokens and one event token.
        self.assertEqual(3, len(workflow_tokens))

    def test_parse_workflows(self):
        """Test parsing WORKFLOWS."""
        from pinball_ext.workflow.parser import PyWorkflowParser
        params =\
            {'workflows_config':
                 'tests.pinball_ext.workflow.parser_test.WORKFLOWS'}
        py_workflow_parser = PyWorkflowParser(params)
        py_workflow_parser.parse_workflows()
        self.assertEquals(3, len(py_workflow_parser.workflows))
        self.assertEquals(2, len(
            py_workflow_parser.workflows['some_workflow'].jobs))
        self.assertEquals(3, len(
            py_workflow_parser.workflows['some_other_workflow'].jobs))
        self.assertEquals(3, len(
            py_workflow_parser.workflows['another_workflow'].jobs))

        self.assertEquals(24 * 60 * 60,
                          py_workflow_parser.workflows['some_workflow'].
                          schedule.recurrence_seconds)
        self.assertEquals(60 * 60,
                          py_workflow_parser.workflows['some_other_workflow'].
                          schedule.recurrence_seconds)
        self.assertEquals(60 * 60,
                          py_workflow_parser.workflows['another_workflow'].
                          schedule.recurrence_seconds)
        self.assertEquals(py_workflow_parser.workflows_config_str,
                          'tests.pinball_ext.workflow.parser_test.WORKFLOWS')

FINAL_JOB_CONFIG = JobConfig(CommandJobTemplate('final', 'true'))
WORKFLOWS = {
    'some_workflow': WorkflowConfig(
        jobs={
            'some_job': JobConfig(
                CommandJobTemplate('some_job', 'true'), [])
        },
        final_job_config=FINAL_JOB_CONFIG,
        schedule=ScheduleConfig(recurrence=timedelta(days=1),
                                reference_timestamp=datetime(
                                    year=2015, month=2, day=1, second=1)),
        notify_emails='foo@pinterest.com'),
    'some_other_workflow': WorkflowConfig(
        jobs={
            'some_other_job': JobConfig(
                CommandJobTemplate('some_other_job', 'true'),
                []),

            'yet_another_job': JobConfig(
                CommandJobTemplate('yet_another_job', 'true'),
                ['some_other_job'])
        },
        final_job_config=FINAL_JOB_CONFIG,
        schedule=ScheduleConfig(recurrence=timedelta(hours=1),
                                reference_timestamp=datetime(
                                    year=2015, month=2, day=1, second=2)),
        notify_emails='foo@pinterest.com'),
    'another_workflow': WorkflowConfig(
        jobs={
            'another_job': JobConfig(
                CommandJobTemplate('another_job', 'true'),
                []),

            'yet_some_other_job': JobConfig(
                CommandJobTemplate('yet_some_other_job', 'true'),
                ['another_job'])
        },
        final_job_config=FINAL_JOB_CONFIG,
        schedule=ScheduleConfig(recurrence=timedelta(hours=1),
                                reference_timestamp=datetime(
                                    year=2015, month=2, day=1, second=2)),
        notify_emails='bar@pinterest.com'),
}



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" The job_runner.py command is responsible for running any *single* Job.

    It supports Bash scripts, Python scripts, Hive scripts and Hadoop jobs.

    It will either be triggered by Pinball or manually via command line and it's
    responsible for parsing and passing along job specific parameters.

    Parameters:
    See _get_cmd_line_options() for supported parameters.

    The required parameters are:
        - job_import_dirs_config: python var name which stores the dirs where
            all job classes are defined.
        - job_class_name: name of the job class we are going to run.

    Users can pass arguments to their Jobs using the --job_params parameter.

    E.g.
    python pinball_ext/job/job_runner.py \
    --job_class_name=SampleReportHiveJob \
    --job_params="start_date=2013-07-27,end_date=2013-07-27" \

    It is up to individual jobs to support date ranges and produce the expected
    output.
"""

import fcntl
import optparse
import traceback

from pinball_ext.common import utils
from pinball_ext.job import job_module


__author__ = 'Changshu Liu, Mohammad Shahangian, Mao Ye'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


LOG = utils.get_logger('pinball_ext.job.job_runner')


def _acquire_exclusive_lock(write_lock_name):
    """Acquires a lock with a given name.

    Underneath we create a lock file and return a descriptor of that file. The
    lock will be held as long as this file descriptor is open.
    """
    lock_filename = '/var/lock/%s.lock' % write_lock_name
    LOG.info('Acquiring lock %s ...', lock_filename)

    lock_file = open(lock_filename, 'w')
    fcntl.flock(lock_file.fileno(), fcntl.LOCK_EX)
    LOG.info('Acquired lock %s.', lock_filename)
    return lock_file


def _get_job_import_dirs(job_import_dirs_config):
    """Get the list of dirs where job classes are defined.

    Args:
        job_import_dirs_config: the variable name that stores a list of dirs
    where job classes are defined.

    NOTE:
        If what user specified is a full qualified variable name, we need to
        import the module first, otherwise, Python won't be able to interpret
        the variable nae correctly.
    """
    if "." in job_import_dirs_config:
        module_name = '.'.join(job_import_dirs_config.split('.')[:-1])
        exec('import %s' % module_name)
    return eval(job_import_dirs_config)


def _get_cmd_line_options():
    parser = optparse.OptionParser()
    parser.add_option("--job_class_name",
                      dest="job_class_name",
                      help="The class name of the job to run.")
    parser.add_option("--dry_run",
                      dest="dry_run",
                      action="store_true",
                      help="If true, don't execute or complete the jobs.")
    parser.add_option("--job_params",
                      dest="job_params",
                      default="",
                      help="These parameters are passed as 'params=' to the job class, and then "
                           "passed along all the way through the job hierarchy. Any arguments "
                           "that user logic might need to access should be put here. Format are "
                           "comma delimited and each have the format '{key}={value}'.")
    parser.add_option("--write_lock_name",
                      dest="write_lock_name",
                      default="",
                      help="Will require an exclusive lock with a given name before running the job. "
                           "Locks are enforced only among jobs running on the same machine. "
                           "The corresponding lock file is named /var/lock/<write_lock_name>.lock")
    parser.add_option("--executor",
                      dest="executor",
                      default="",
                      help="The name of the main executor for the job. The list of settings is in "
                           "pinball_ext/executor/common.py.")
    parser.add_option("--executor_config",
                      dest="executor_config",
                      default="",
                      help="These settings are passed along to all of the Executor constructors. "
                           "They are for overriding the default Executor configuration settings. "
                           "Settings are comma delimited and each have the format '{key}={value}'.")
    parser.add_option("--job_import_dirs_config",
                      dest="job_import_dirs_config",
                      default="",
                      help="Full qualified Python variable name which stores the list of dirs that "
                           "contains job definition files.")

    (options, args) = parser.parse_args()
    return parser, options, args


def main():
    parser, opts, args = _get_cmd_line_options()

    # Load job Python class
    assert opts.job_import_dirs_config, '--job_import_dirs_config not set.'
    job_import_dirs = _get_job_import_dirs(opts.job_import_dirs_config)
    LOG.info('loading job classes from: %s', '\n\t'.join(job_import_dirs))

    job_class_name = opts.job_class_name
    job_py_class = job_module.get_py_class_by_name(job_import_dirs,
                                                   job_class_name)
    if not job_py_class:
        all_class_names = '\n\t' + '\n\t'.join(
            job_module.get_sorted_job_names(job_import_dirs))
        LOG.error('job_class_name:%s not found, it should be one of:%s',
                  job_class_name,
                  all_class_names)
        raise Exception('Python class with name: %s not found' % job_class_name)

    # Executor related settings.
    settings = {}
    if opts.executor:
        settings['executor'] = opts.executor
    executor_config = utils.parse_arguments(opts.executor_config)
    settings['executor_config'] = executor_config

    # Config job parameters
    params = utils.parse_arguments(opts.job_params)

    # Run the job
    lock_file = None
    if opts.write_lock_name:
        lock_file = _acquire_exclusive_lock(opts.write_lock_name)

    try:
        LOG.info("starts to run job: %s on executor: %s ...",
                 job_class_name,
                 opts.executor)

        # Create the job object
        job_object = job_py_class(params=params, settings=settings)
        job_object.runjob(dry_run=opts.dry_run)
    except Exception:
        LOG.warning("Error: %s", traceback.format_exc())
        LOG.error("job failed.")
        raise
    finally:
        if lock_file:
            lock_file.close()

    LOG.info("job succeeded.")


if __name__ == "__main__":
    main()



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import mock
from pinball_ext.executor.common import Platform
from pinball_ext.job import hive_jobs


__author__ = 'Changshu Liu'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'

_Q_TEMPLATE = """SELECT * FROM unknown_db.unknown_tb WHERE dt >= '%(start_date)s' AND dt <= '%(end_date)s'"""

_PARAMS = {
    'start_date': '2015-01-01',
    'end_date': '2015-01-07',
}

_SETTINGS = {
    'executor': Platform.EMR,
    'executor_config': {
        'USER_APPJAR_PATH': 's3n:/my_bucket/my_dir/my_jar.jar',
    }
}

_JOB_STDOUT = 'stdout'
_JOB_STDERR = 'stderr'
_JOB_IDS = ['9527']


class HiveTestJob(hive_jobs.HiveJob):
    _QUERY_TEMPLATE = _Q_TEMPLATE


class HiveFileTestJob(hive_jobs.HiveFileJob):
    # Check HiveFileJob to see where the dir starts.
    _QUERY_TEMPLATE_FILE = '../../tests/pinball_ext/job/hive_test_query.ql'


def _mock_emr_run_hive_query_job(q_str, upload_archive):
    return _JOB_STDOUT, _JOB_STDERR, _JOB_IDS


class HiveJobsTestCase(unittest.TestCase):
    @mock.patch('pinball_ext.job.basic_jobs.ClusterJob._is_prod_job',
                mock.Mock(return_value=False))
    @mock.patch('pinball_ext.executor.emr_executor.EMRExecutor.__init__',
                mock.Mock(return_value=None))
    @mock.patch('pinball_ext.executor.emr_executor.EMRExecutor.run_hive_query',
                side_effect=_mock_emr_run_hive_query_job)
    def test_hive_query_job(self, f1):

        hive_job = HiveTestJob(params=_PARAMS, settings=_SETTINGS)
        hive_job.runjob()
        self.assertEquals(hive_job._job_output, _JOB_STDOUT)
        self.assertEquals(hive_job._job_stderr, _JOB_STDERR)
        self.assertEquals(hive_job._job_ids, _JOB_IDS)
        f1.assert_called_once_with((_Q_TEMPLATE % _PARAMS),
                                   upload_archive=False)

    @mock.patch('pinball_ext.job.basic_jobs.ClusterJob._is_prod_job',
                mock.Mock(return_value=False))
    @mock.patch('pinball_ext.executor.emr_executor.EMRExecutor.__init__',
                mock.Mock(return_value=None))
    @mock.patch('pinball_ext.executor.emr_executor.EMRExecutor.run_hive_query',
                side_effect=_mock_emr_run_hive_query_job)
    def test_hive_file_job(self, f1):
        hive_job = HiveFileTestJob(params=_PARAMS, settings=_SETTINGS)
        hive_job.runjob()
        self.assertEquals(hive_job._job_output, _JOB_STDOUT)
        self.assertEquals(hive_job._job_stderr, _JOB_STDERR)
        self.assertEquals(hive_job._job_ids, _JOB_IDS)
        f1.assert_called_once_with((_Q_TEMPLATE % _PARAMS),
                                   upload_archive=False)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Run tasks at predefined time intervals."""
import pickle
import time

from pinball.config.utils import PinballException
from pinball.config.utils import get_log
from pinball.config.utils import get_unique_name
from pinball.config.utils import token_to_str

from pinball.master.thrift_lib.ttypes import ModifyRequest
from pinball.master.thrift_lib.ttypes import Query
from pinball.master.thrift_lib.ttypes import QueryAndOwnRequest
from pinball.master.thrift_lib.ttypes import TokenMasterException
from pinball.scheduler.overrun_policy import OverrunPolicy
from pinball.workflow.name import Name


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


LOG = get_log('pinball.scheduler.scheduler')


class Scheduler(object):
    # How long to own the schedule token while manipulating it.
    _LEASE_TIME_SEC = 5 * 60  # 5 minutes
    # How long to sleep after no un-owned schedule token has been found.
    _SLEEP_TIME_SEC = 10
    # How long to delay the schedule if it's already running and appropriate
    # policy is in place.
    _DELAY_TIME_SEC = 5 * 60  # 5 minutes

    def __init__(self, client, store, emailer):
        self._client = client
        self._store = store
        self._emailer = emailer
        self._owned_schedule_token = None
        self._request = None
        self._name = get_unique_name()
        self._test_only_end_if_no_unowned = False

    def _own_schedule_token(self):
        """Attempt to own a schedule token.

        Try to own a schedule token.  Only unowned tokens will be considered.
        Unowned schedules are ready to run.  The ownership of the qualifying
        job token lasts for a limited time so it has to be periodically renewed
        if the schedule takes longer than that to run.
        """
        assert not self._owned_schedule_token
        query = Query()
        query.namePrefix = Name.SCHEDULE_PREFIX
        query.maxTokens = 1
        request = QueryAndOwnRequest()
        request.query = query
        request.expirationTime = int(time.time()) + Scheduler._LEASE_TIME_SEC
        request.owner = self._name
        try:
            response = self._client.query_and_own(request)
            if response.tokens:
                assert len(response.tokens) == 1
                self._owned_schedule_token = response.tokens[0]
        except TokenMasterException:
            LOG.exception('')

    def _advance_schedule(self, schedule):
        schedule.advance_next_run_time()
        self._owned_schedule_token.expirationTime = schedule.next_run_time
        self._owned_schedule_token.data = pickle.dumps(schedule)

    def _abort_workflow(self, schedule):
        return schedule.abort_running(self._client, self._store)

    def _run_or_reschedule(self):
        """Run the schedule represented by the owned schedule token.

        If the time is right and the overrun policy permits it, run the owned
        schedule token.  Otherwise, reschedule it until a later time.
        """
        assert self._owned_schedule_token
        schedule = pickle.loads(self._owned_schedule_token.data)
        if schedule.next_run_time > time.time():
            # It's not time to run it yet.  Although we should claim only
            # tokens which are ready to run, clock skew between different
            # machines may result in claiming a token too soon.
            assert (self._owned_schedule_token.expirationTime >=
                    schedule.next_run_time), ('%d < %d in token %s' % (
                        self._owned_schedule_token.expirationTime,
                        schedule.next_run_time,
                        token_to_str(self._owned_schedule_token)))
        elif (schedule.overrun_policy == OverrunPolicy.START_NEW or
              schedule.overrun_policy == OverrunPolicy.ABORT_RUNNING or
              # Ordering of the checks in the "and" condition below is
              # important to avoid a race condition when a workflow gets
              # retried and changes the state from failed to running.
              ((schedule.overrun_policy != OverrunPolicy.DELAY_UNTIL_SUCCESS or
                not schedule.is_failed(self._store)) and
               not schedule.is_running(self._store))):
            if schedule.overrun_policy == OverrunPolicy.ABORT_RUNNING:
                if not self._abort_workflow(schedule):
                    return
            self._request = schedule.run(self._emailer, self._store)
            if self._request:
                self._advance_schedule(schedule)
        elif schedule.overrun_policy == OverrunPolicy.SKIP:
            self._advance_schedule(schedule)
        elif (schedule.overrun_policy == OverrunPolicy.DELAY or
              schedule.overrun_policy == OverrunPolicy.DELAY_UNTIL_SUCCESS):
            self._owned_schedule_token.expirationTime = int(
                time.time() + Scheduler._DELAY_TIME_SEC)
        else:
            raise PinballException('unknown schedule policy %d in token %s' % (
                schedule.overrun_policy, self._owned_schedule_token))

    def _update_tokens(self):
        """Update tokens modified during schedule execution in the master.
        """
        assert self._owned_schedule_token
        if not self._request:
            self._request = ModifyRequest()
        if not self._request.updates:
            self._request.updates = []
        self._request.updates.append(self._owned_schedule_token)
        schedule = pickle.loads(self._owned_schedule_token.data)
        if schedule.workflow == 'experiments':
            LOG.info('updating tokens for workflow experiments %s',
                     self._request)
        try:
            self._client.modify(self._request)
        except TokenMasterException:
            LOG.exception('')
        finally:
            self._owned_schedule_token = None
            self._request = None

    def run(self):
        """Run the scheduler."""
        LOG.info('Running scheduler ' + self._name)
        while True:
            self._own_schedule_token()
            if self._owned_schedule_token:
                self._run_or_reschedule()
                self._update_tokens()
            elif self._test_only_end_if_no_unowned:
                return
            else:
                time.sleep(Scheduler._SLEEP_TIME_SEC)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generic workflow-related utilities."""
import importlib
import os
import time

from pinball.config.pinball_config import PinballConfig


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


def get_unique_workflow_instance():
    """Generate a unique workflow instance name.

    Returns:
        A timestamp-based workflow instance name.
    """
    # We don't expect instance names to be generated for the same workflow
    # at sub-millisecond intervals so timestamp should be good enough.
    return str(int(time.time() * 1000))


def load_path(path):
    """Load an object identified by a fully qualified python path.

    Args:
        path: The path of the object to load.
    Returns:
        The object identified by the path.
    """
    index = path.rfind('.')
    if index == -1:
        return None
    module_name = path[:index]
    attribute_name = path[(index + 1):]
    module = importlib.import_module(module_name)
    return getattr(reload(module), attribute_name)


def get_logs_dir(workflow, instance,
                 parent_log_directory=PinballConfig.LOCAL_LOGS_DIR):
    """Get name of the directory where workflow instance logs are stored.

    Args:
        workflow: Name of the workflow whose logs are stored in the returned
            directory.
        instance: Name of the workflow instance whose logs are stored in the
            returned directory.
        parent_log_directory: Name of the parent directory where workflow instance
            logs are stored.
    Returns:
        Name of the directory where workflow instance logs are stored.
    """
    return os.path.join(parent_log_directory, workflow, instance)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Command line tool to start token master and workflow workers."""
import argparse
import gc
import guppy
import signal
import socket
import sys
import time
import threading

from django.core import management
from pinball.config.pinball_config import PinballConfig
from pinball.config.utils import get_log
from pinball.config.utils import master_name
from pinball.master.factory import Factory
from pinball.ui import cache_thread


__author__ = 'Pawel Garbacki, Mao Ye'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = ['Pawel Garbacki', 'Mao Ye']
__license__ = 'Apache'
__version__ = '2.0'


LOG = get_log('pinball.run_pinball')

DbStore = None
Scheduler = None
Emailer = None
Worker = None


def _pinball_imports():
    """Import Pinball modules.

    The reason why these imports are not at the top level is that some of the
    imported code (db models initializing table names) depends on parameters
    passed on the command line (master name).  Those imports need to be delayed
    until after command line parameter parsing.
    """

    global DbStore
    from pinball.persistence.store import DbStore
    assert DbStore

    global Scheduler
    from pinball.scheduler.scheduler import Scheduler
    assert Scheduler

    global Emailer
    from pinball.workflow.emailer import Emailer
    assert Emailer

    global Worker
    from pinball.workflow.worker import Worker
    assert Worker


def _create_scheduler(factory, emailer):
    result = threading.Thread(target=_run_scheduler, args=[factory, emailer])
    result.daemon = True
    result.start()
    return result


def _run_scheduler(factory, emailer):
    client = factory.get_client()
    scheduler = Scheduler(client, DbStore(), emailer)
    scheduler.run()


def _create_workers(num_workers, factory, emailer):
    threads = []
    # A delay between starting individual workers.  We space starting new
    # workers to prevent overwhelming the master.  The interval should be long
    # enough to give a worker time to connect to the master but not so long
    # that it would substantially delay processing.
    sleep_interval = 5. / num_workers
    sleep_interval = max(sleep_interval, 1)
    for _ in range(0, num_workers):
        thread = threading.Thread(target=_run_worker, args=[factory, emailer])
        thread.daemon = True
        threads.append(thread)
        thread.start()
        time.sleep(sleep_interval)
    return threads


def _run_worker(factory, emailer):
    client = factory.get_client()
    worker = Worker(client, DbStore(), emailer)
    worker.run()


def _wait_for_threads(threads):
    finished_threads = 0
    total_threads = len(threads)
    LOG.info('Waiting for %d thread(s) to finish' % total_threads)
    # We cannot simply go over the list of threads and join them one by one
    # because join() does not propagate KeyboardInterrupt.
    while threads:
        new_threads = []
        found_finished = False
        for thread in threads:
            if thread.isAlive():
                new_threads.append(thread)
            else:
                found_finished = True
                thread.join()
                finished_threads += 1
                LOG.info('Thread %d/%d finished' % (finished_threads,
                                                    total_threads))
        threads = new_threads
        if not found_finished:
            time.sleep(5)
    LOG.info('Exiting')
    sys.exit()


def _siguser1_handler(sig, frame):
    """Signal handler showing some memory info.

    Args:
        sig: The captured signal.
        frame: The current stack frame.
    """
    gc.collect()
    heapy = guppy.hpy()
    heap = heapy.heap()
    for i in range(0, 5):
        LOG.info(heap)
        heap = heap.more
    relative_memory = heapy.heap().get_rp(50)
    for i in range(0, 5):
        LOG.info(relative_memory)
        relative_memory = relative_memory.more
    heapy.setref()


def _register_signal_listener():
    """Set up a signal handler showing memory stats."""
    signal.signal(signal.SIGUSR1, _siguser1_handler)


def main():
    _register_signal_listener()

    parser = argparse.ArgumentParser(
        description='Start Pinball master and workers.')
    parser.add_argument(
        '-c',
        '--config_file',
        dest='config_file',
        required=True,
        help='full path to the pinball setting configure file')
    parser.add_argument(
        '-m',
        '--mode',
        dest='mode',
        choices=['master', 'scheduler', 'workers', 'ui'],
        default='master',
        help='execution mode')

    options = parser.parse_args(sys.argv[1:])
    PinballConfig.parse(options.config_file)

    if hasattr(PinballConfig, 'MASTER_NAME') and PinballConfig.MASTER_NAME:
        master_name(PinballConfig.MASTER_NAME)
    _pinball_imports()
    if PinballConfig.UI_HOST:
        emailer = Emailer(PinballConfig.UI_HOST, PinballConfig.UI_PORT)
    else:
        emailer = Emailer(socket.gethostname(), PinballConfig.UI_PORT)

    if options.mode == 'ui':
        hostport = '%s:%d' % (socket.gethostname(), PinballConfig.UI_PORT)
        cache_thread.start_cache_thread(DbStore())
        if not PinballConfig.UI_HOST:
            hostport = 'localhost:%d' % PinballConfig.UI_PORT

        # Disable reloader to prevent auto refresh on file changes.  The
        # problem with auto-refresh is that it starts multiple processes.  Some
        # of those processes will become orphans if we kill the UI in a wrong
        # way.
        management.call_command('runserver', hostport, interactive=False,
                                use_reloader=False)
        return

    factory = Factory(master_hostname=PinballConfig.MASTER_HOST,
                      master_port=PinballConfig.MASTER_PORT)
    threads = []
    if options.mode == 'master':
        factory.create_master(DbStore())
    elif options.mode == 'scheduler':
        threads.append(_create_scheduler(factory, emailer))
    else:
        assert options.mode == 'workers'
        if PinballConfig.UI_HOST:
            emailer = Emailer(PinballConfig.UI_HOST, PinballConfig.UI_PORT)
        else:
            emailer = Emailer(socket.gethostname(), PinballConfig.UI_PORT)
        threads = _create_workers(PinballConfig.WORKERS, factory, emailer)

    try:
        if options.mode == 'master':
            factory.run_master_server()
        else:
            _wait_for_threads(threads)
    except KeyboardInterrupt:
        LOG.info('Exiting')
        sys.exit()


if __name__ == '__main__':
    main()



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import getpass
import os
import resource
import subprocess
import time

from pinball_ext.common import hadoop_utils
from pinball_ext.common import utils
from pinball_ext.executor.common import make_executor
from pinball_ext.executor.common import Platform


__author__ = 'Mao Ye, Mohammad Shahangian, Changshu Liu'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


_SUCCESS_FILE = '_SUCCESS'
JOB_SLEEP_TIME_SEC = 60
LOG = utils.get_logger('pinball_ext.job.basic_jobs')


class JobBase(object):
    """Base class for all actual Jobs.

    All actual Jobs will take a (possibly empty) dictionary of parameters that
    it will parse and apply to its calculation.

    args:
        params - key value pairs for the params that the job will run using

    To run a Job we invoke the runjob() method which decides how to call
    the following methods:
        - _setup(): Responsible for constructing the arguments for the command
            line tool being invoked and doing any other prep work such as
            compilation.

        _set_output_dirs(): Set the attribute self.job_output_dirs with list of
            local or s3 directories for output files. These directories will be
            deleted when the job fails if _CLEANUP_AFTER_FAILURE is set to True.

        _execute(): Responsible for running the executable with the parameters
            that were setup(). Some jobs will have more complex execute methods,
            e.g. running the executable once for each day.

            This method should raise an exception if the job could not execute
            successfully.

            If this job produced any output that needs to be persisted or
            post-processed, the _execute function is also responsible for
            assigning that data to the "self._job_output" variable. The output
            is expected to be in list-of-lists format, which can roughly be
            translated into row-column format:
            self._job_output = [
                ['cell1a', 'cell1b'], # row 1
                ['cell2a', 'cell2b'], # row 2
            ]

            The _execute function is also responsible for assigning the
            "self._job_stderr" variable which takes the same format of data as
            _job_output.

        _complete(): Responsible for persisting any of the query results.
            For example, the _REPORTER is invoked in the _complete method.

        _cleanup(): Cleans up any temporary local files, or temporary S3/HDFS
            files. This will be called even if the other methods raise
            exceptions.
    """
    # If set to True, output dynamically generated during a failed job will
    # be deleted. Subclass jobs extending JobBase should override this to True
    # if needed. Do not change this default value.
    _CLEANUP_AFTER_FAILURE = False

    # If set to true, skip the job if the output dir has the _SUCCESS file.
    _SKIP_IF_SUCCEEDED = False

    def __init__(self, params=None, settings=None):
        """
        Args:
            params - a dict config that might be used by job logic.
            settings - a dict config that's used to execute job.
        """
        self.params = params if params else {}
        self.settings = settings if settings else {}

        self.user = getpass.getuser()
        job_name = self.params.get('job_name', None)
        self.job_name = job_name if job_name is not None else self.__class__.__name__
        self.delay_execution = 0

        # This is set in _set_output_dirs function
        self.job_output_dirs = []
        # These fields are set by the _execute function
        self._job_output = None
        self._job_stderr = None

    def _skip_execution(self):
        if self._SKIP_IF_SUCCEEDED and hasattr(self, 'output'):
            success_file = os.path.join(self.output, _SUCCESS_FILE)
            if hadoop_utils.hdfs_exists(success_file):
                LOG.info('Found {0}. Skip this job.'.format(success_file))
                return True
        return False

    def runjob(self, dry_run=False):
        """The runjob method is responsible for calling the setup, execute
        and complete for the dates that the report will be run.
        """
        try:
            self._setup()
            self._set_output_dirs()
            if dry_run:
                print self
                return
            if not self._skip_execution():
                self._execute()
            self._complete()
        except Exception:
            LOG.exception("Failed to run job")
            if self._CLEANUP_AFTER_FAILURE:
                self._cleanup_output_dirs(self.job_output_dirs)
            raise
        finally:
            self._cleanup()
        return

    def _setup(self):
        pass

    def _execute(self):
        pass

    def _complete(self):
        LOG.info('Maximum Resident set was %d MB' % (
            resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1024))
        LOG.info('Successfully ran %s' % self.job_name)

    def _cleanup(self):
        pass

    def _cleanup_output_dirs(self, directories):
        LOG.info("Job failed: Output generated during job will get deleted.")
        for directory in directories:
            if self._is_local_directory(directory):
                self._delete_local_directory(directory)
            else:
                hadoop_utils.hdfs_rmr(directory)

    @staticmethod
    def _is_local_directory(directory):
        """Args:
            directory: A string representation of the directory path
        """
        return os.path.isdir(directory)

    @staticmethod
    def _delete_local_directory(directory):
        if directory[-1] != '/':
            directory += '/'
        proc = subprocess.Popen(
            'rm -rf %s' % directory,
            shell=True, stdout=subprocess.PIPE)
        assert proc.communicate()[0] is ''

    def _get_input_output_resources(self):
        # Override in child job classes.
        pass

    def _set_output_dirs(self):
        # Override in child job classes.
        pass

    def __str__(self):
        return '(%s): (%s)' % (self.job_name, self.params)


class PythonJob(JobBase):
    """PythonJob execute python code.

    The code should exist in the same codebase that this class lives. Each job
    is responsible for importing the dependencies in its execute() method.
    """
    def _setup(self):
        LOG.info('Warning: This PythonJob doesn\'t have any setup step.')

    def _execute(self):
        NotImplementedError('PythonJob does not have an execute method')


class CommandLineJob(JobBase):
    """CommandLineJob run a command line command.

    The user will usually need to override the setup method to format the
    arguments in the format that its command expects.
    """
    def _setup(self):
        self.arguments = ''
        for k, v in self.params.iteritems():
            self.arguments += '--%s=%s ' % (k, v)

    def _execute(self):
        command = '%s %s' % (self._get_command(), self.arguments)
        LOG.info('Running command: %s' % command)
        proc = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE)

        # Get command process stdout.
        self._job_output = proc.communicate()[0]
        proc.wait()
        if proc.returncode != 0:
            raise Exception('Command %s failed' % command)

    def _get_command(self):
        raise Exception('This job does not have a command associated with it')


class ClusterJob(JobBase):
    def __init__(self, params, settings=None):
        super(ClusterJob, self).__init__(params, settings=settings)

        self._counters = {}
        self._job_ids = []

        if 'executor' not in self.settings:
            self.settings['executor'] = Platform.EMR
            LOG.warning(
                "%s is missing an 'executor' parameter in the job_settings. "
                "Using the default executor: %s.",
                self.job_name,
                self.settings['executor'])
        executor_name = self.settings['executor']
        executor_config = self.settings.get('executor_config', {})
        executor_config['NAME'] = self.job_name
        executor_config['PLATFORM'] = executor_name
        self.executor = make_executor(executor_name, executor_config)

    @property
    def hadoop_job_name(self):
        """The name of the job as it appears on the Hadoop Job Tracker."""
        return self.executor.job_name

    # on the mutable variable self._job_ids. rename to get_counters()
    @property
    def counters(self):
        """Return the job counters for the jobs in this Cluster job.

        Returns dict from job ID to a map of counter names to values for each
        job created by the Cluster job.
        """
        if not self._counters:
            for jid in self._job_ids:
                self._counters[jid] = self.executor.get_hadoop_counters(jid)
        return self._counters

    def _is_prod_job(self):
        return self.executor.config.USER == 'prod'

    def _delay(self):
        """Sleep for a while.

        This method is invoked by operations that depend on s3 files which may
        not be available due to s3 eventual consistency issues.
        """
        if self._is_prod_job():
            LOG.info('waiting for %d seconds to give s3 time to replicate',
                     JOB_SLEEP_TIME_SEC)
            time.sleep(JOB_SLEEP_TIME_SEC)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of the token master logic."""
import pytrie
import sys
import threading

from pinball.config.utils import get_log
from pinball.master.blessed_version import BlessedVersion
from pinball.master.transaction import REQUEST_TO_TRANSACTION


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


LOG = get_log('pinball.master.master_handler')


class MasterHandler(object):
    """Handler implementing the token master logic.

    Tokens are stored in a trie where keys are token names while the values are
    the tokens themselves.  Trie structure provides an efficient access to
    operations on token name prefixes such as token querying and counting.

    A special type of singleton token - called the blessed version - is stored
    in the tree with other tokens.  The blessed version is used to generate
    unique version numbers.
    """
    _BLESSED_VERSION = '/__BLESSED_VERSION__'
    _MASTER_OWNER = '__master__'

    def __init__(self, store):
        self._store = store
        self._trie = pytrie.StringTrie()
        self._lock = threading.Lock()
        self._load_tokens()

    def _load_tokens(self):
        try:
            tokens = self._store.read_active_tokens()
            for token in tokens:
                self._trie[token.name] = token
            blessed_version = self._trie.get(MasterHandler._BLESSED_VERSION)
            if blessed_version:
                # Note that blessed_version is an instance of Token class.  We
                # need a BlessedVersion object.
                self._trie[MasterHandler._BLESSED_VERSION] = (
                    BlessedVersion.from_token(blessed_version))
            else:
                assert not self._trie
                self._trie[MasterHandler._BLESSED_VERSION] = BlessedVersion(
                    MasterHandler._BLESSED_VERSION,
                    MasterHandler._MASTER_OWNER)
                self._store.commit_tokens(
                    updates=[self._trie[MasterHandler._BLESSED_VERSION]])
        except:
            LOG.exception('')
            # A failure here may indicate that the token tree was populated
            # partially.  It's dangerous to continue in this state so just
            # exit.
            sys.exit(1)

    def _process_request(self, request):
        transaction_cls = REQUEST_TO_TRANSACTION[request.__class__]
        transaction = transaction_cls()
        transaction.prepare(request)
        with self._lock:
            # TODO(pawel): it would be cleaner to subclass trie implementing
            # auto-persistence in the store when modifying tokens.
            return transaction.commit(self._trie,
                                      self._trie[
                                          MasterHandler._BLESSED_VERSION],
                                      self._store)

    # TODO(pawel): add a meta-operation inferring what to do from the class
    # of the request.

    def archive(self, request):
        return self._process_request(request)

    def group(self, request):
        return self._process_request(request)

    def modify(self, request):
        return self._process_request(request)

    def query(self, request):
        return self._process_request(request)

    def query_and_own(self, request):
        return self._process_request(request)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A token container supporting persistence."""
import abc

from pinball.config.utils import set_django_environment
set_django_environment()

from django import db
from django.core import management
from django.db import transaction

from pinball.config.utils import get_log
from pinball.persistence.models import ActiveTokenModel
from pinball.persistence.models import ArchivedTokenModel
from pinball.persistence.models import CachedDataModel


__author__ = 'Pawel Garbacki, Mao Ye'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = ['Pawel Garbacki', 'Mao Ye']
__license__ = 'Apache'
__version__ = '2.0'


LOG = get_log('pinball.persistence.store')


class Store(object):
    """An interface for persistent token containers."""
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        self.initialize()

    @abc.abstractmethod
    def initialize(self):
        """Initialize the token store."""
        return

    @abc.abstractmethod
    def commit_tokens(self, updates=None, deletes=None):
        """Update or remove active tokens.

        Args:
            updates: The list of active token updates to persist.  Some tokens
            on this list may not exist in the store in which case they get
                inserted.
            deletes: The list of active tokens to remove from the store.
                Tokens on this list are required to exist in the store.
        """
        return

    def delete_archived_tokens(self, deletes):
        """Remove archived tokens.

        Args:
            deletes: The list of archived tokens to remove from the store.
                Tokens on this list are required to exist in the store.
        """
        return

    @abc.abstractmethod
    def read_active_tokens(self, name_prefix='', name_infix='',
                           name_suffix=''):
        """Read active tokens with names matching the provided filters."""
        return

    @abc.abstractmethod
    def read_archived_tokens(self, name_prefix='',
                             name_infix='',
                             name_suffix=''):
        """Read archived tokens with names matching the provided filters."""
        return

    @abc.abstractmethod
    def archive_tokens(self, tokens):
        """Move active tokens to archived store."""
        return

    @abc.abstractmethod
    def get_cached_data(self, name):
        """Get data with a given name from the cache."""
        return

    @abc.abstractmethod
    def set_cached_data(self, name, data):
        """Set data with a given name in the cache."""
        return

    @abc.abstractmethod
    def read_tokens(self, name_prefix='', name_infix='', name_suffix=''):
        """Read tokens with names matching the provided filters."""
        return

    @abc.abstractmethod
    def read_token_names(self, name_prefix='', name_infix='', name_suffix=''):
        """Read token names matching the provided filters."""
        return

    @abc.abstractmethod
    def read_archived_token_names(self, name_prefix='', name_infix='',
                                  name_suffix=''):
        """Read archived token names matching the provided filters."""
        return

    @abc.abstractmethod
    def read_cached_data_names(self, name_prefix='', name_infix='',
                               name_suffix=''):
        """Read cached data names matching the provided filters."""
        return


class DbStore(Store):
    """Implementation of token store on top of a database."""
    def initialize(self):
        """Create db tables if they don't exist."""
        management.call_command('syncdb', interactive=False)

    @transaction.commit_on_success
    def commit_tokens(self, updates=None, deletes=None):
        updates = updates if updates is not None else []
        deletes = deletes if deletes is not None else []
        for token in updates:
            token_model = ActiveTokenModel.from_token(token)
            token_model.save()
        for token in deletes:
            token_model = ActiveTokenModel.from_token(token)
            token_model.delete()

    @transaction.commit_on_success
    def delete_archived_tokens(self, deletes):
        for token in deletes:
            token_model = ArchivedTokenModel.from_token(token)
            token_model.delete()

    def read_active_tokens(self, name_prefix='', name_infix='',
                           name_suffix=''):
        # Refresh the connection to make sure we read the most recent snapshot.
        db.close_connection()
        return self._read_active_tokens(name_prefix, name_infix, name_suffix)

    @transaction.commit_on_success
    def _read_active_tokens(self, name_prefix='', name_infix='',
                            name_suffix=''):
        result = []
        token_models = self._read_data(ActiveTokenModel, name_prefix,
                                       name_infix, name_suffix)
        for token_model in token_models:
            result.append(token_model.to_token())
        return result

    def read_archived_tokens(self, name_prefix='', name_infix='',
                             name_suffix=''):
        db.close_connection()
        return self._read_archived_tokens(name_prefix, name_infix, name_suffix)

    @transaction.commit_on_success
    def _read_archived_tokens(self, name_prefix='', name_infix='',
                              name_suffix=''):
        result = []
        token_models = self._read_data(ArchivedTokenModel, name_prefix,
                                       name_infix, name_suffix)
        for token_model in token_models:
            result.append(token_model.to_token())
        return result

    def get_cached_data(self, name):
        db.close_connection()
        return self._get_cached_data(name)

    @transaction.commit_on_success
    def _get_cached_data(self, name):
        try:
            return CachedDataModel.objects.get(name=name).data
        except CachedDataModel.DoesNotExist:
            return None

    @transaction.commit_on_success
    def set_cached_data(self, name, data):
        cached_data = CachedDataModel.from_data(name, data)
        cached_data.save()

    @transaction.commit_on_success
    def archive_tokens(self, tokens):
        for token in tokens:
            active_token_model = ActiveTokenModel.from_token(token)
            active_token_model.delete()
            archived_token_model = ArchivedTokenModel.from_token(token)
            archived_token_model.save()

    def read_tokens(self, name_prefix='', name_infix='', name_suffix=''):
        db.close_connection()
        return self._read_tokens(name_prefix, name_infix, name_suffix)

    @transaction.commit_on_success
    def _read_tokens(self, name_prefix='', name_infix='', name_suffix=''):
        result = []
        active_tokens = self._read_data(ActiveTokenModel, name_prefix,
                                        name_infix, name_suffix)
        for token in active_tokens:
            result.append(token)
        archived_tokens = self._read_data(ArchivedTokenModel, name_prefix,
                                          name_infix, name_suffix)
        for token in archived_tokens:
            result.append(token)
        return result

    def read_token_names(self, name_prefix='', name_infix='', name_suffix=''):
        db.close_connection()
        return self._read_token_names(name_prefix, name_infix, name_suffix)

    def _read_token_names(self, name_prefix='', name_infix='', name_suffix=''):
        result = []
        active_token_names = self._read_data(ActiveTokenModel, name_prefix,
                                             name_infix, name_suffix, True)
        result.extend(active_token_names)
        archived_token_names = self._read_data(ArchivedTokenModel, name_prefix,
                                               name_infix, name_suffix, True)
        result.extend(archived_token_names)
        return result

    def read_archived_token_names(self, name_prefix='', name_infix='',
                                  name_suffix=''):
        db.close_connection()
        return self._read_archived_token_names(name_prefix, name_infix,
                                               name_suffix)

    def _read_archived_token_names(self, name_prefix='', name_infix='',
                                   name_suffix=''):
        result = []
        archived_token_names = self._read_data(ArchivedTokenModel, name_prefix,
                                               name_infix, name_suffix, True)
        result.extend(archived_token_names)
        return result

    def read_cached_data_names(self, name_prefix='', name_infix='',
                               name_suffix=''):
        db.close_connection()
        return self._read_cached_data_names(name_prefix, name_infix,
                                            name_suffix)

    def _read_cached_data_names(self, name_prefix='', name_infix='',
                                name_suffix=''):
        result = []
        cached_token_names = self._read_data(CachedDataModel, name_prefix,
                                             name_infix, name_suffix, True)
        result.extend(cached_token_names)
        return result

    @transaction.commit_on_success
    def clear_cached_data(self):
        CachedDataModel.objects.all().delete()

    def _customized_name_filter_query(self, model, name_prefix='', name_infix='',
                                      name_suffix='', select_name_only=False):
        """ Customize query based on given name filtering criteria.

        Instead of django.db filter operator, we use customized query is
        due to the following issue. django.db always translates string
        filter query by using "LIKE BINARY". Unfortunately, LIKE BINARY
        does not respect the index for the field with varchar data type.

        Args:
            model: django db model that contains table's information
            name_prefix: name starts with this string
            name_infix: name contains this string
            name_suffix: name ends with this string
            select_name_only: only return name in the query result

        Returns:
            Customized query
        """
        query_template = 'SELECT %s FROM %s %s;'
        selector = 'name' if select_name_only else '*'
        where_condition = ''
        conditions = []
        if name_prefix != '':
            conditions.append('name LIKE "%s%%%%"' % name_prefix.replace('_', '\_'))
        if name_infix != '':
            conditions.append('name LIKE "%%%%%s%%%%"' % name_infix.replace('_', '\_'))
        if name_suffix != '':
            conditions.append('name LIKE "%%%%%s"' % name_suffix.replace('_', '\_'))
        if conditions:
            where_condition = 'WHERE ' + ' AND '.join(conditions)

        query = query_template % (selector, model.get_table_name(), where_condition)
        return query

    @transaction.commit_on_success
    def _read_data(self, model, name_prefix='', name_infix='',
                   name_suffix='', select_name_only=False):
        raw_query = self._customized_name_filter_query(model, name_prefix,
                                                       name_infix, name_suffix,
                                                       select_name_only)
        results = model.objects.raw(raw_query)
        if select_name_only:
            # Unicode conversion extracts the value of the 'name' field
            # only from the model object.
            results = [unicode(result) for result in results]
        return results



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Validation tests for log saver"""

import unittest

from pinball.config.pinball_config import PinballConfig
from pinball.workflow import log_saver


__author__ = 'Mao Ye'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class FileLogSaverTestCase(unittest.TestCase):
    def setUp(self):
        self._local_file_path = '/tmp/local/file/path'
        self._log_saver = log_saver.FileLogSaver.from_path(self._local_file_path)

    def test_from_path(self):
        isinstance(self._log_saver, log_saver.FileLogSaver)


class S3FileLogSaverTestCase(unittest.TestCase):
    def setUp(self):
        PinballConfig.S3_LOGS_DIR_PREFIX = 's3n://bucket_name/'
        self._s3_path = 's3n://bucket_name/pinball_job_logs/rest_file_path'
        self._log_saver = log_saver.FileLogSaver.from_path(self._s3_path)

    def test_from_path(self):
        isinstance(self._log_saver, log_saver.S3FileLogSaver)

    def test_local_file_log_saver(self):
        isinstance(self._log_saver._local_file_log_saver, log_saver.FileLogSaver)
        self.assertEqual(self._log_saver._local_file_log_saver._file_path,
                         PinballConfig.LOCAL_LOGS_DIR+'rest_file_path')

# TODO(mao): Add more comprehensive tests including
# 1) data is read-from and written-to correct paths,
# 2) s3 log saver creates non-conflicting local file paths,
# 3) writing to s3 log saver delays the upload based on time and size.



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Constructs shared by multiple tools."""
import abc

from pinball.config.utils import PinballException


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class CommandException(PinballException):
    pass


class Command(object):
    """Interface for commands interacting with the master."""
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def prepare(self, options):
        """Parse command options."""
        return

    @abc.abstractmethod
    def execute(self, client, store):
        """Execute the command using a provided client.

        Args:
            client: The client connected to the master.
            store: The token store.
        Returns:
            Output of the command.  We return a string rather than printing it
            to stdout to make testing easier - mocking 'print' is a pain.
        """
        # TODO(pawel): modify this method to return a tuple: (exit_code,
        # stdout, stderr)
        return


def confirm(prompt='Confirm'):
    """Prompt for yes or no response from the user.

    Args:
        prompt: A message to show when asking for confirmation.
    Returns:
        True iff the user approved the action.
    """
    prompt = '%s [%s]|%s: ' % (prompt, 'n', 'y')

    while True:
        ans = raw_input(prompt)
        if not ans:
            return False
        if ans not in ['y', 'Y', 'n', 'N']:
            print 'please enter y or n.'
            continue
        if ans == 'y' or ans == 'Y':
            return True
        if ans == 'n' or ans == 'N':
            return False
